# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import numpy
import bisect

from datetime import timedelta
from datetime import datetime
from datetime import date
from dateutil.relativedelta import relativedelta
import cftime


def convert_date_to_datetime(mydate):
    mydatetime = datetime.combine(mydate, datetime.min.time())
    return mydatetime


def convert_datetime_to_date(mydatetime):
    mydate = mydatetime.date()
    return mydate


def get_start_of_the_day(dt):
    dt0 = datetime(dt.year, dt.month, dt.day)
    return dt0


def get_end_of_the_day(dt):
    dt0 = datetime(dt.year, dt.month, dt.day, 23, 59, 59)
    return dt0


def convert_datevec_to_datetime(datevec):
    order = ['year', 'month', 'day', 'hour', 'minute', 'second']
    dt_str = "{:04d}".format(datevec[0]) + "{:02d}".format(datevec[1]) + \
             "{:02d}".format(datevec[2]) + "{:02d}".format(datevec[3]) + \
             "{:02d}".format(datevec[4]) + "{:02d}".format(datevec[5])
    dt0 = datetime.strptime(dt_str, "%Y%m%d%H%M%S")
    return dt0


def convert_datetime_to_datevec(dt0):
    dt_vec = [dt0.year, dt0.month, dt0.day, dt0.hour, dt0.minute, dt0.second]
    return dt_vec


def convert_datetime_to_sectime(dts, dt0=None):
    dts_type = type(dts)
    if dts_type is datetime:
        dts = [dts]

    if isinstance(dts, (list, tuple)):
        dts = numpy.array(dts)
    shape = dts.shape

    if dt0 is None:
        dt0 = get_start_of_the_day(dts.flatten()[0])

    dt_delta = dts.flatten() - dt0
    sectime = [dt_temp.total_seconds() for dt_temp in dt_delta]

    if dts_type is list:
        return sectime, dt0
    elif dts_type is tuple:
        return tuple(sectime), dt0
    else:
        sectime = numpy.array(sectime).reshape(shape)
        return sectime, dt0


def get_diff_months(dt1, dt2):
    diff_months = ((dt2.year - dt1.year) * 12) + dt2.month - dt1.month
    return diff_months


def get_diff_days(dt1, dt2):
    dtdelta = datetime(dt2.year, dt2.month, dt2.day) - datetime(dt1.year, dt1.month, dt1.day)
    return int(numpy.ceil(dtdelta.total_seconds() / 86400))


def get_first_day_of_month(dt):
    if type(dt) is date:
        mydate = dt
        datetype = 0
    elif issubclass(dt.__class__, datetime):
        mydate = convert_datetime_to_date(dt)
        datetype = 1
    this_month = mydate - timedelta(days=mydate.day - 1)
    if datetype == 0:
        return this_month
    if datetype == 1:
        return convert_date_to_datetime(this_month)


def get_last_day_of_month(dt, end=False):
    if type(dt) is date:
        mydate = dt
        datetype = 0
    elif issubclass(dt.__class__, datetime):
        mydate = convert_datetime_to_date(dt)
        datetype = 1
    next_month = mydate.replace(day=28) + timedelta(days=4)  # this will never fail
    end_month = next_month - timedelta(days=next_month.day)

    if datetype == 0:
        return end_month
    if datetype == 1:
        dt = convert_date_to_datetime(end_month)
        if end:
            dt = dt + timedelta(seconds=86400-1)
        return dt


def get_next_of_month(dt):
    if type(dt) is date:
        mydate = dt
        datetype = 0
    elif issubclass(dt.__class__, datetime):
        mydate = convert_datetime_to_date(dt)
        datetype = 1
    next_month = mydate.replace(day=28) + timedelta(days=4)  # this will never fail
    next_month = next_month - timedelta(days=next_month.day - 1)
    if datetype == 0:
        return next_month
    if datetype == 1:
        return convert_date_to_datetime(next_month)


def get_next_n_months(dt, nm):
    if type(dt) is date:
        mydate = dt
        datetype = 0
    elif issubclass(dt.__class__, datetime):
        mydate = convert_datetime_to_date(dt)
        datetype = 1
    this_month = get_first_day_of_month(mydate)
    next_n_month = this_month + relativedelta(months=nm)
    if datetype == 0:
        return next_n_month
    if datetype == 1:
        return convert_date_to_datetime(next_n_month)


def convert_datetime_to_matlabdn(dts):
    type_in = type(dts)
    dts = numpy.array(dts)
    dns = numpy.empty_like(dts)
    for ind, dt in enumerate(dts.flatten()):

        mdn = dt + timedelta(days=366)
        frac_seconds = (dt - datetime(dt.year, dt.month, dt.day, 0, 0, 0)).seconds / (24.0 * 60.0 * 60.0)
        frac_microseconds = dt.microsecond / (24.0 * 60.0 * 60.0 * 1000000.0)
        dns[ind] = mdn.toordinal() + frac_seconds + frac_microseconds
    if type_in is datetime:
        return dns[0]
    elif type_in is list:
        return dns.tolist()
    elif 'numpy' in str(type_in):
        return dns.astype(numpy.double)

def convert_matlabdn_to_datetime(dns):
    type_in = type(dns)
    dns = numpy.array(dns)
    dts = numpy.empty_like(dns, dtype=datetime)
    for ind, dn in enumerate(dns.flatten()):
        dts[ind] = datetime.fromordinal(int(dn)) + timedelta(days=dn % 1) - timedelta(days=366)

    if type_in is datetime:
        return dts[0]
    elif type_in is list:
        return dts.tolist()
    elif 'numpy' in str(type_in):
        return dts


def get_doy(dts, year=None, decimal=False):
    type_in = type(dts)
    if type_in is datetime:
        dts = [dts]
    dts = numpy.array(dts)
    doys = numpy.empty_like(dts)

    if year is not None:
        dt0 = datetime(year, 1, 1, 0, 0, 0)
    else:
        dt0 = datetime(dts[0].year, 1, 1, 0, 0, 0)

    for ind, dt in enumerate(dts.flatten()):
        delta_dt = dt - dt0
        doys[ind] = delta_dt.total_seconds()/86400. + 1
    if not decimal:
        doys = numpy.floor(doys)
    if type_in is datetime:
        return doys[0]
    elif type_in is list:
        return doys.tolist()
    elif 'numpy' in str(type_in):
        return doys


def convert_doy_to_datetime(year, doys):
    type_in = type(doys)
    if isinstance(doys, (int, float)):
        doys = [doys]
    doys = numpy.array(doys)
    dts = numpy.empty_like(doys, dtype=datetime)
    dt0 = datetime(year, 1, 1, 0, 0, 0)
    total_seconds = (doys - 1) * 86400.
    for ind, total_second in enumerate(total_seconds.flatten()):
        delta_dt = timedelta(seconds=total_second)
        dts[ind] = dt0 + delta_dt

    if type_in in (int, float):
        return dts[0]
    elif type_in is list:
        return dts.tolist()
    elif 'numpy' in str(type_in):
        return dts


def convert_unix_time_to_datetime(times):
    type_in = type(times)

    ts = numpy.array(times)
    dts = numpy.empty_like(ts, dtype=datetime)

    for ind, t in enumerate(ts.flatten()):
        dts[ind] = datetime.utcfromtimestamp(t)

    if type_in in (int, float):
        return dts[0]
    elif type_in is list:
        return dts.tolist()
    elif 'numpy' in str(type_in):
        return dts


def convert_unix_time_to_datetime_cftime(times):
    type_in = type(times)

    ts = numpy.array(times)
    dts = numpy.empty_like(ts, dtype=datetime)

    dts = numpy.array(
        cftime.num2date(ts.flatten(),
                        units='seconds since 1970-01-01 00:00:00.0',
                        only_use_cftime_datetimes=False,
                        only_use_python_datetimes=True)
    )

    if type_in in (int, float):
        return dts[0]
    elif type_in is list:
        return dts.tolist()
    elif 'numpy' in str(type_in):
        return dts.reshape(times.shape)


def convert_gps_time_to_datetime(times, weeks=None):
    
    type_in = type(times)

    ts = numpy.array(times).flatten()
    if weeks is None:
        weeks = numpy.zeros_like(ts)
    else:
        weeks = numpy.array(weeks).flatten()

    gps_seconds = weeks * _SECONDS_PER_WEEK + times
    gps_seconds_add = numpy.array([bisect.bisect_left(_LEAP_SECONDS_GPS_TIME, gs) for gs in gps_seconds])
    dts = numpy.array([_GPS_DATETIME_0 + timedelta(seconds=sec) for sec in (gps_seconds - gps_seconds_add)])

    if type_in in (int, float):
        return dts[0]
    elif type_in is list:
        return dts.tolist()
    elif 'numpy' in str(type_in):
        return dts.reshape(times.shape)

def convert_datetime_to_gps_times(times: datetime, with_weeks=False):
    
    type_in = type(times)

    ts = numpy.array(times).flatten()
    
    ts_seconds = numpy.array([ (t - _GPS_DATETIME_0).total_seconds() for t in ts])
    print(ts_seconds)
    
    leap_seconds_dates = [datetime(i[0], i[1], i[2], 23, 59, 59) for i in _LEAP_SECONDS_DATES_GPS] 

    gps_seconds = numpy.array([sec + bisect.bisect_left(leap_seconds_dates, t) for sec, t in zip(ts_seconds, ts)])
    print(gps_seconds)
    if with_weeks:
        weeks = numpy.floor(gps_seconds / _SECONDS_PER_WEEK)
        gps_seconds = gps_seconds % _SECONDS_PER_WEEK
        if type_in is datetime:
            return gps_seconds[0], weeks[0]
        elif type_in is list: 
            return gps_seconds.tolist(), weeks.tolist()
        elif 'numpy' in str(type_in):
            return gps_seconds.reshape(times.shape), weeks.reshape(times.shape)
    else:
        if type_in is datetime:
            return gps_seconds[0]
        elif type_in is list: 
            return gps_seconds.tolist()
        elif 'numpy' in str(type_in):
            return gps_seconds.reshape(times.shape)
        

_GPS_DATETIME_0 = datetime(1980, 1, 6)
_SECONDS_PER_WEEK = 604800.0
_LEAP_SECONDS_DATES = [
    (1972, 6, 30), (1972, 12, 31), (1973, 12, 31), (1974, 12, 31), (1975, 12, 31), 
    (1976, 12, 31), (1977, 12, 31), (1978, 12, 31), (1979, 12, 31), (1980, 6, 30), 
    (1982, 6, 30), (1983, 6, 30), (1985, 6, 30), (1987, 12, 31), (1989, 12, 31), 
    (1990, 12, 31), (1992, 6, 30), (1993, 6, 30),  (1994, 6, 30), (1995, 12, 31),
    (1997, 6, 30), (1998, 12, 31), (2005, 12, 31), (2008, 12, 31), (2012, 6, 30),
    (2015, 6, 30), (2016, 12, 31), 
]

_LEAP_SECONDS_DATES_GPS = [
    (1980, 6, 30), (1982, 6, 30), (1983, 6, 30), (1985, 6, 30), (1987, 12, 31), 
    (1989, 12, 31), (1990, 12, 31), (1992, 6, 30), (1993, 6, 30),  (1994, 6, 30), 
    (1995, 12, 31), (1997, 6, 30), (1998, 12, 31), (2005, 12, 31), (2008, 12, 31), 
    (2012, 6, 30), (2015, 6, 30), (2016, 12, 31), 
]

_LEAP_SECONDS_COUNT_GPS = [
    0., 1., 2., 3., 4., 
    5., 6., 7., 8., 9., 10., 
    11., 12., 13., 14., 15.,
    16., 17., 18., 
]

# gps_seconds = [(datetime(i[0], i[1], i[2], 23, 59, 59) - _GPS_DATETIME_0).total_seconds() + j - 1 for i, j in zip(_LEAP_SECONDS_DATES, _LEAP_SECONDS_COUNT_GPS)]

_LEAP_SECONDS_GPS_TIME = [
    15292799.0, 78364800.0, 109900801.0, 173059202.0, 252028803.0, 
    315187204.0, 346723205.0, 393984006.0, 425520007.0, 457056008.0, 
    504489609.0, 551750410.0, 599184011.0, 820108812.0, 914803213.0, 
    1025136014.0, 1119744015.0, 1167264016.0
]




import numpy as np
from scipy.signal import argrelextrema
import datetime
from scipy.interpolate import griddata
from scipy.signal import butter, lfilter, freqz
import scipy.signal as sig

from geospacelab.datahub import DatasetUser
from geospacelab.toolbox.utilities import pydatetime as dttool
from geospacelab.cs import GEOCSpherical


class LEOToolbox(DatasetUser):

    def __init__(self, dt_fr=None, dt_to=None):
        super().__init__(dt_fr, dt_to)

        self.add_variable(var_name='SC_GEO_LAT')
        self.add_variable(var_name='SC_GEO_LON')
        self.add_variable(var_name='SC_GEO_ALT')
        self.add_variable(var_name='SC_GEO_LST')
        self.add_variable(var_name='SC_DATETIME')
        self.add_variable(var_name='SC_AACGM_LAT')
        self.add_variable(var_name='SC_AACGM_LON')
        self.add_variable(var_name='SC_AACGM_MLT')
        self.add_variable(var_name='SC_APEX_LAT')
        self.add_variable(var_name='SC_APEX_LON')
        self.add_variable(var_name='SC_APEX_MLT')
        self.ascending_nodes = {}
        self.descending_nodes = {}
        self.northern_nodes = {}
        self.southern_nodes = {}
        self.sector_cs = 'GEO'
        self.sectors = {}

    def search_orbit_nodes(self):
        glat = self['SC_GEO_LAT'].value.flatten()
        glon = self['SC_GEO_LON'].value.flatten()
        lst = self['SC_GEO_LST'].value.flatten()
        dts = self['SC_DATETIME'].value.flatten()
        ind_1 = argrelextrema(np.abs(glat), np.less)[0]
        ind_2 = argrelextrema(glat, np.greater)[0]
        ind_3 = argrelextrema(glat, np.less)[0]
        inds_asc = []
        inds_dsc = []
        for ind_N in ind_2:
            if ind_N < ind_1[0]:
                inds_dsc.append(ind_1[0])
            elif ind_N > ind_1[-1]:
                inds_asc.append(ind_1[-1])
            else:
                ind_between = np.where(ind_1 > ind_N)[0]
                inds_asc.append(ind_1[ind_between[0]-1])
                inds_dsc.append(ind_1[ind_between[0]])
        self.ascending_nodes['INDEX'] = inds_asc
        self.ascending_nodes['GEO_LON'] = glon[inds_asc]
        self.ascending_nodes['GEO_LST'] = lst[inds_asc]
        self.ascending_nodes['DATETIME'] = dts[inds_asc]
        self.descending_nodes['INDEX']= inds_dsc
        self.descending_nodes['GEO_LON'] = glon[inds_dsc]
        self.descending_nodes['GEO_LST'] = lst[inds_dsc]
        self.descending_nodes['DATETIME'] = dts[inds_asc]

        self.northern_nodes['INDEX'] = ind_2
        self.northern_nodes['GEO_LON'] = glon[ind_2]
        self.northern_nodes['GEO_LST'] = lst[ind_2]
        self.northern_nodes['DATETIME'] = dts[ind_2]
        self.southern_nodes['INDEX'] = ind_3
        self.southern_nodes['GEO_LON'] = glon[ind_3]
        self.southern_nodes['GEO_LST'] = lst[ind_3]
        self.southern_nodes['DATETIME'] = dts[ind_3]


    def group_by_sector(self, sector_name, boundary_lat, sector_cs='GEO'):
        self.sector_cs = sector_cs
        if not dict(self.ascending_nodes):
            self.search_orbit_nodes()

        glat = self['SC_GEO_LAT'].value.flatten()
        glon = self['SC_GEO_LON'].value.flatten()
        lst = self['SC_GEO_LST'].value.flatten()
        
        if self.sector_cs == 'AACGM':
            mlat = self['SC_AACGM_LAT'].value.flatten()
            mlon = self['SC_AACGM_LON'].value.flatten()
            mlt = self['SC_AACGM_MLT'].value.flatten()
        elif self.sector_cs == 'APEX':
            mlat = self['SC_APEX_LAT'].value.flatten()
            mlon = self['SC_APEX_LON'].value.flatten()
            mlt = self['SC_APEX_MLT'].value.flatten()
        
        dts = self['SC_DATETIME'].value.flatten()
        
        inds_asc = self.ascending_nodes['INDEX']
        inds_dsc = self.descending_nodes['INDEX']

        if self.sector_cs == 'GEO':
            lat = glat
            lt = lst
        elif self.sector_cs == 'AACGM':
            lat = mlat
            lt = mlt
        elif self.sector_cs == 'APEX':
            lat = mlat
            lt = mlt
        else:
            raise NotImplementedError

        inds_center_N = argrelextrema(lat, np.greater)[0]
        # inds_center_N = inds_center_N[np.where(lat[inds_center_N] > boundary_lat)[0]]

        inds_center_S = argrelextrema(lat, np.less)[0]
        # inds_center_S = inds_center_S[np.where(lat[inds_center_S] < -np.abs(boundary_lat))[0]]

        # Sector centered at the northern pole from the ascending node towards the descending node.
        if sector_name == 'N':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_center_N):

                ind_tmp = np.where(inds_center_S < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S1 = inds_center_S[ind_tmp[-1]]
                else:
                    ind_rec_S1 = 0
                ind_tmp = np.where(inds_center_S > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S2 = inds_center_S[ind_tmp[0]]
                else:
                    ind_rec_S2 = len(lat) - 1
                inds_seg = np.array(range(ind_rec_S1, ind_rec_S2))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((lat_seg > boundary_lat) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c

                lat_seg = np.where(dts_seg < dt_c, lat_seg, 180. - lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_N', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_N_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_N_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['N'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [boundary_lat, 180. - boundary_lat],
                'VARIABLE_NAMES': ['SECTOR_N', 'SECTOR_N_DATETIME', 'SECTOR_N_PSEUDO_LAT']
            }
        
        # Sector centered at the southern pole from the descending node towards the ascending node.
        elif sector_name == 'S':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_center_S):

                ind_tmp = np.where(inds_center_N < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N1 = inds_center_N[ind_tmp[-1]]
                else:
                    ind_rec_N1 = 0
                ind_tmp = np.where(inds_center_N > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N2 = inds_center_N[ind_tmp[0]]
                else:
                    ind_rec_N2 = len(lat) - 1
                inds_seg = np.array(range(ind_rec_N1, ind_rec_N2))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((lat_seg < -np.abs(boundary_lat)) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c

                lat_seg = np.where(dts_seg < dt_c, 180. - lat_seg, 360. + lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_S', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_S_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_S_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['S'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [180. + np.abs(boundary_lat), 360. - np.abs(boundary_lat)],
                'VARIABLE_NAMES': ['SECTOR_S', 'SECTOR_S_DATETIME', 'SECTOR_S_PSEUDO_LAT']
            }

        # Sector centered at the asending node from south towards north.
        elif sector_name == 'ASC':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_asc):

                ind_tmp = np.where(inds_center_S < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S1 = inds_center_S[ind_tmp[-1]]
                else:
                    ind_rec_S1 = 0
                ind_tmp = np.where(inds_center_N > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N1 = inds_center_N[ind_tmp[0]]
                else:
                    ind_rec_N1 = len(lat) - 1 
                inds_seg = np.array(range(ind_rec_S1, ind_rec_N1))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((np.abs(lat_seg) < np.abs(boundary_lat)) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c

                lat_seg = np.where(lat_seg < 0, lat_seg, lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_ASC', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_ASC_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_ASC_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['ASC'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [-np.abs(boundary_lat), np.abs(boundary_lat)],
                'VARIABLE_NAMES': ['SECTOR_ASC', 'SECTOR_ASC_DATETIME', 'SECTOR_ASC_PSEUDO_LAT']
            }
            
            # Sector centered at the asending node from south towards north.
        elif sector_name == 'DSC':
            sector = np.zeros((dts.size,))
            dts_c = np.zeros_like(dts)

            pseudo_lat = np.empty_like(sector)
            pseudo_lat[:] = np.nan
            for num, ind_rec in enumerate(inds_dsc):

                ind_tmp = np.where(inds_center_N < ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_N1 = inds_center_N[ind_tmp[-1]]
                else:
                    ind_rec_N1 = 0
                ind_tmp = np.where(inds_center_S > ind_rec)[0]
                if list(ind_tmp):
                    ind_rec_S1 = inds_center_S[ind_tmp[0]]
                else:
                    ind_rec_S1 = len(lat) - 1
                inds_seg = np.array(range(ind_rec_N1, ind_rec_S1))

                dt_c = dts[ind_rec]
                lat_seg = lat[inds_seg]
                dts_seg = dts[inds_seg]
                delta_t_seg = np.array([(dt1 - dt_c).total_seconds() / 60. for dt1 in dts_seg])
                inds_seg_seg = np.where((np.abs(lat_seg) < np.abs(boundary_lat)) & (np.abs(delta_t_seg) < 60.))[0]

                inds_sector = inds_seg[inds_seg_seg]
                sector[inds_sector] = num + 1
                dts_c[inds_sector] = dt_c

                lat_seg = np.where(lat_seg > 0, 180. - lat_seg, 180. - lat_seg)
                pseudo_lat[inds_sector] = lat_seg[inds_seg_seg]

            self.add_variable(var_name='SECTOR_DSC', value=sector.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_DSC_DATETIME', value=dts_c.reshape((sector.size, 1)))
            self.add_variable(var_name='SECTOR_DSC_PSEUDO_LAT', value=pseudo_lat.reshape((sector.size, 1)))
            self.sectors['DSC'] = {
                'BOUNDARY_LAT': boundary_lat,
                'PSEUDO_LAT_RANGE': [180. - np.abs(boundary_lat), 180. + np.abs(boundary_lat)],
                'VARIABLE_NAMES': ['SECTOR_DSC', 'SECTOR_DSC_DATETIME', 'SECTOR_DSC_PSEUDO_LAT']
            }
        else:
            raise NotImplementedError
                
    def griddata_by_sector(self, sector_name=None, variable_names=None, x_grid_res=20*60, y_grid_res=0.5):
        self.visual = 'on'
        dts_c = self['_'.join(('SECTOR', sector_name, 'DATETIME'))].value.flatten()
        dts = self['SC_DATETIME'].value.flatten()
        sector = self['SECTOR_' + sector_name].value.flatten()
        lat = self['_'.join(('SECTOR', sector_name, 'PSEUDO_LAT'))].value.flatten()
        boundary_lat = self.sectors[sector_name]['BOUNDARY_LAT']
        lat_range = self.sectors[sector_name]['PSEUDO_LAT_RANGE']

        x_data = dts_c[sector > 0]
        y_data = lat[sector > 0]
        is_finite = np.isfinite(y_data)
        dt0 = dttool.get_start_of_the_day(self.dt_fr)
        sectime = np.array([(dt - dt0).total_seconds() for dt in x_data])
        x = sectime
        y = y_data
        # max_x = np.ceil(np.max(x) / self.xgrid_res) * self.xgrid_res
        # min_x = np.floor(np.min(x) / self.xgrid_res) * self.xgrid_res
        total_seconds = (dts[-1] - dt0).total_seconds()
        min_x = 0
        max_x = total_seconds

        nx = int(np.ceil((max_x - min_x) / x_grid_res) + 1)
        ny = int(np.ceil(np.diff(lat_range) / y_grid_res) + 1)

        grid_x, grid_y = np.meshgrid(np.linspace(min_x, max_x, nx),
                                    np.linspace(lat_range[0], lat_range[1], ny))

        grid_lat = griddata((x, y), y, (grid_x, grid_y), method='nearest')
        if self.sector_cs == 'AACGM':
            mask = np.abs(grid_y - grid_lat) > 2
            which_lat = 'AACGM_LAT'
        if self.sector_cs == 'APEX':
            mask = np.abs(grid_y - grid_lat) > 2
            which_lat = 'APEX_LAT' 
        else:
            mask = np.abs(grid_y - grid_lat) > 1
            which_lat = 'GEO_LAT'

        method = 'linear'
        for var_name in variable_names:
            vrb = self[var_name].value.flatten()[sector>0]
            grid_z = griddata((x, y), vrb, (grid_x, grid_y), method=method)
            grid_z[mask] = np.nan
            var_name_in = '_'.join(('SECTOR', sector_name, 'GRID', var_name))
            self.sectors[sector_name]['VARIABLE_NAMES'].append(var_name_in) 
            
            # self.add_variable(var_name=var_name_in, value=grid_z.T)
            self[var_name_in] = self[var_name].clone(omit_attrs='visual')
            self[var_name_in].visual = 'new'
            var = self[var_name_in]
            var.value = grid_z.T
            var.set_depend(0, {'UT': '_'.join(('SECTOR', sector_name, 'GRID_DATETIME'))})
            var.set_depend(1, {which_lat: '_'.join(('SECTOR', sector_name, 'GRID_LAT'))})
            var.visual.axis[0].data = '@d.' + '_'.join(('SECTOR', sector_name, 'GRID_DATETIME'))
            var.visual.axis[1].data = '@d.' + '_'.join(('SECTOR', sector_name, 'GRID_LAT'))
            var.visual.axis[1].label = 'GLAT' if self.sector_cs == 'GEO' else 'MLAT'
            var.visual.axis[1].unit = r'$^\circ$'
            var.visual.axis[1].lim = lat_range
            var.visual.axis[2].data = '@v.value'
            var.visual.axis[2].data_scale = self[var_name].visual.axis[1].data_scale
            var.visual.axis[2].label = '@v.label'
            var.visual.axis[2].unit = '@v.unit'
            var.visual.plot_config.style = '2P'
            var.visual.plot_config.pcolormesh.update(cmap='jet')

        grid_x_name = '_'.join(('SECTOR', sector_name, 'GRID', 'X'))
        grid_x = grid_x.T
        self.add_variable(var_name=grid_x_name, value=grid_x)
        grid_y_name = '_'.join(('SECTOR', sector_name, 'GRID', 'Y'))
        grid_y = grid_y.T
        self.add_variable(var_name=grid_y_name, value=grid_y)
        var_name = '_'.join(('SECTOR', sector_name, 'GRID_DATETIME'))
        grid_dts = np.array([dt0 + datetime.timedelta(seconds=sec) for sec in grid_x[:, 0]])
        self.add_variable(var_name=var_name, value=grid_dts[:, np.newaxis])
        var_name = '_'.join(('SECTOR', sector_name, 'GRID_LAT'))
        self.add_variable(var_name=var_name, value=grid_y[0, :][np.newaxis, :])

    def filter_by_time(self, variable_names=None, time_window=40., time_res=None):


        return

    @staticmethod
    def format_pseudo_lat_label(ax, sector_name, is_integer=True):
        if is_integer:
            lat_format = '{:4.0f}'
        else:
            lat_format = '{:5.1f}'
        y_lim = ax.get_ylim()
        y_max = np.max(y_lim)
        y_min = np.min(y_lim)
        yticks = ax.get_yticks()
        ylabels = []
        n1 = 0
        n2 = len(yticks) - 1
        n3 = 0
        for ind, pseudo_lat in enumerate(yticks):
            if pseudo_lat < y_min:
                n1 = n1 + 1
                ylabels.append('')
                continue
            if pseudo_lat > y_max:
                n2 = n2 - 1
                ylabels.append('')
                continue
            if sector_name == 'N':
                lat = pseudo_lat if pseudo_lat <= 90. else 180. - pseudo_lat
                if lat == 90.:
                    n3 = ind

            elif sector_name == 'S':
                lat = 180. - pseudo_lat if pseudo_lat <= 270. else pseudo_lat - 360.
                if lat == -90:
                    n3 = ind

            elif sector_name == 'ASC':
                lat = pseudo_lat
                if lat == 0.:
                    n3 = ind

            elif sector_name == 'DSC':
                lat = 180. - pseudo_lat
                if lat == 0.:
                    n3 = ind

            else:
                raise NotImplementedError
            if lat > 0:
                lat_format_1 = r'' + lat_format + r'$^\circ$N'
                lb = lat_format_1.format(lat)
            elif lat < 0:
                lat_format_1 = r'' + lat_format + r'$^\circ$S'
                lb = lat_format_1.format(np.abs(lat))
            else:
                lat_format_1 = r'' + '{:4.0f}' + r'$^\circ$'
                lb = lat_format_1.format(lat) 
            ylabels.append(lb)

        if sector_name == 'N':
            ylabels[n1] = 'ASC:' + ylabels[n1]
            ylabels[n2] = 'DSC:' + ylabels[n2]
            ylabels[n3] = 'Pole->' + ylabels[n3]
        elif sector_name == 'S':
            ylabels[n1] = 'DSC:' + ylabels[n1]
            ylabels[n2] = 'ASC:' + ylabels[n2]
            ylabels[n3] = 'Pole->' + ylabels[n3]
        elif sector_name == 'ASC':
            ylabels[n1] = '' + ylabels[n1]
            ylabels[n2] = '' + ylabels[n2]
            ylabels[n3] = 'ASC->' + ylabels[n3]
        elif sector_name == 'DSC':
            ylabels[n1] = '' + ylabels[n1]
            ylabels[n2] = '' + ylabels[n2]
            ylabels[n3] = 'DSC->' + ylabels[n3]
        else:
            raise NotImplementedError
        ax.set_yticks(yticks)
        ax.set_yticklabels(ylabels)
        ax.set_ylim(y_lim)

        return

    def smooth_along_track(self, time_window=450, time_res=None, variable_names=None, method='savgol', residuals='on'):
        self.visual = 'on'
        dts = self['SC_DATETIME'].value.flatten()
        dt0 = dttool.get_start_of_the_day(self.dt_fr)
        sectime = np.array([(dt - dt0).total_seconds() for dt in dts])
        if time_res is None:
            diff_secs = np.diff(sectime)
            time_res = np.median(diff_secs)

        N = int(np.fix(time_window/time_res))

        for var_name in variable_names:
            vrb = self[var_name].value.flatten()
            vrb_avg = self.smooth_savgol(dts, vrb, N, time_res)
            var_name_in = var_name + '_AVG'
            self[var_name_in] = self[var_name].clone()
            self[var_name_in].value = vrb_avg[:, np.newaxis]
            self[var_name_in].label = 'smoothed ' + self[var_name_in].label
            if residuals:
                var_name_in = var_name + '_RSD'
                self[var_name_in] = self[var_name].clone()
                self[var_name_in].value = self[var_name].value - self[var_name + '_AVG'].value
                self[var_name_in].label = 'RSD'
                self[var_name_in].visual.axis[1].lim = [-np.inf, np.inf]
                var_name_in = var_name + '_RSD_PERC'
                self[var_name_in] = self[var_name].clone()
                self[var_name_in].value = (self[var_name].value - self[var_name + '_AVG'].value) / \
                    self[var_name + '_AVG'].value
                self[var_name_in].label = "RSD pct."
                self[var_name_in].unit = '%'
                self[var_name_in].unit_label = None
                self[var_name_in].visual.axis[1].data_scale = 100.

    @staticmethod
    def smooth_savgol(dts, data, box_pts, t_res):
        delta_sectime = 1
        sectime, _ = dttool.convert_datetime_to_sectime(dts)
        ind_nan = np.isnan(data)
        sectime_i = np.arange(0, sectime[-1], delta_sectime)
        data_i = np.interp(sectime_i, sectime[~ind_nan], data[~ind_nan])

        order = 1
        window_size = int(np.floor(t_res / delta_sectime * box_pts / 2) * 2 + 1)
        y = sig.savgol_filter(data_i, window_size, order)
        data_new = np.interp(sectime, sectime_i, y)
        data_new[ind_nan] = np.nan
        return data_new
    
    def trajectory_local_unit_vector(self):
        lat_0 = self['SC_GEO_LAT'].flatten()
        lon_0 = self['SC_GEO_LON'].flatten()
        height_0 = self['SC_GEO_ALT'].flatten()
        
        cs_0 = GEOCSpherical(coords={'lat': lat_0, 'lon': lon_0, 'height': height_0})
        phi_0 = cs_0['phi']
        theta_0 = cs_0['theta']
        cs_0 = cs_0.to_cartesian()
        x_0 = cs_0['x']
        y_0 = cs_0['y']
        z_0 = cs_0['z']
        
        dx = np.concatenate(
            (
                [cs_0['x'][1] - cs_0['x'][0]], 
                cs_0['x'][2:] - cs_0['x'][0:-2],
                [cs_0['x'][-1] - cs_0['x'][-2]]
            ),
            axis=0
        )
        dy = np.concatenate(
            (
                [cs_0['y'][1] - cs_0['y'][0]], 
                cs_0['y'][2:] - cs_0['y'][0:-2],
                [cs_0['y'][-1] - cs_0['y'][-2]]
            ),
            axis=0
        )
        dz = np.concatenate(
            (
                [cs_0['z'][1] - cs_0['z'][0]], 
                cs_0['z'][2:] - cs_0['z'][0:-2],
                [cs_0['z'][-1] - cs_0['z'][-2]]
            ),
            axis=0
        )
        
        dv = np.array([dx.flatten(), dy.flatten(), dz.flatten()]).T
        
        v_new = np.empty_like(dv)
        
        for ind, (phi_c, theta_c) in enumerate(zip(phi_0, theta_0)):
            R_1 = np.array([
                [-np.sin(phi_c), -np.cos(phi_c), 0],
                [np.cos(phi_c), -np.sin(phi_c), 0],
                [0, 0, 1]
            ])

            R_2 = np.array([
                [1, 0, 0],
                [0, np.cos(theta_c), -np.sin(theta_c)],
                [0, np.sin(theta_c), np.cos(theta_c)]
            ])
            v_new[ind, :] = dv[ind, :].reshape((1, 3)) @ R_1 @ R_2
            
        
        norm = np.sqrt(v_new[:, 0]**2 + v_new[:, 1]**2 + v_new[:, 2]**2)
        
        v_unit = np.empty_like(v_new)
        v_unit[:, 0] = v_new[:, 1] / norm
        v_unit[:, 1] = v_new[:, 0] / norm
        v_unit[:, 2] = - v_new[:, 2] / norm
        
        return v_unit
        
        
         


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import copy
import numpy as np
from geospacelab.datahub import VariableModel as Var

database = 'OMNI'

timestamps = {
    'DATETIME': 'DATETIME',
}

depend_0 = {'UT': 'DATETIME'}

default_colormap = "gist_ncar"

default_plot_config = {
    'line':         {
        'linestyle':        '-',
        'linewidth':        1.5,
        'marker':           '',
        'markersize':       3,
    },
    'pcolormesh':   {
        'cmap':            default_colormap,
    }
}

configured_variables = {}
visual = 'on'

####################################################################################################################
var_name = 'B_x_GSM'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Magnetic flux density along the GSM x-axis'
var.label = r'$B_x^{GSM}$'
var.group = r'$B$'
var.unit = 'nT'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-np.inf, np.inf]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit_label'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'B_y_GSM'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Magnetic flux density along the GSM y-axis'
var.label = r'$B_y^{GSM}$'
var.group = r'$B$'
var.unit = 'nT'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-np.inf, np.inf]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit_label'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'B_z_GSM'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Magnetic flux density along the GSM z-axis'
var.label = r'$B_z^{GSM}$'
var.group = r'$B$'
var.unit = 'nT'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-np.inf, np.inf]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit_label'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'B_T_GSM'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Transverse magnetic flux density in the GSM y-z plane'
var.label = r'$B_T^{GSM}$'
var.group = r'$B$'
var.unit = 'nT'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-np.inf, np.inf]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit_label'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'B_TOTAL'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Total magnetic flux density'
var.label = r'$B_{TOT}$'
var.group = r'$B$'
var.unit = 'nT'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-np.inf, np.inf]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit_label'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_sw'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Solar wind speed'
var.label = r'$v_{SW}$'
var.group = 'v'
var.unit = 'km/s'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [200, None]
axis[1].label = '@v.label'
axis[1].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'n_p'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Proton density'
var.label = r'$n_\mathrm{p}$'
var.group = r'$n$'
var.unit = 'n/cc'
var.unit_label = r'cm$^{-3}$'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [0.5, 50]
axis[1].scale = 'log'
axis[1].label = '@v.group'
axis[1].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'p_dyn'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'dynamic pressure'
var.label = r'$p_\mathrm{dyn}$'
var.group = r'$p$'
var.unit = 'nPa'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [None, None]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit_label'

configured_variables[var_name] = var


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import os
import madrigalWeb.madrigalWeb as madrigalweb
import geospacelab.toolbox.utilities.pybasic as pybasic
import geospacelab.toolbox.utilities.pylogging as mylog

default_madrigal_url = "http://cedar.openmadrigal.org/"

"""
Functions for the Madrigal sources
"""


def list_all_instruments(madrigal_url=default_madrigal_url, database=None):
    # list all the instruments from the madrigal sources
    # get sources info
    if database is None:
        database = madrigalweb.MadrigalData(madrigal_url)

    # List all instruments
    inst_list = database.getAllInstruments()
    mylog.simpleinfo.info("List all the instruments from the Madrigal sources:\n")
    for inst in inst_list:
        mylog.simpleinfo.info("%s: %s", str(inst.code), inst.name)


def list_experiments(instrument_code, dt_fr, dt_to, madrigal_url=default_madrigal_url, database=None):
    if database is None:
        database = madrigalweb.MadrigalData(madrigal_url)

    exp_list = database.getExperiments(
        instrument_code,
        dt_fr.year, dt_fr.month, dt_fr.day, dt_fr.hour, dt_fr.minute, dt_fr.second,
        dt_to.year, dt_to.month, dt_to.day, dt_to.hour, dt_to.minute, dt_to.second,
        local=0
    )

    if not list(exp_list):
        raise ValueError('Cannot find the experiments for the database! Check the input values!')

    if exp_list[0].id == -1:
        madrigal_url = exp_list[0].madrigalUrl
        mylog.simpleinfo.info("Madrigal sources has been relocated to %s", madrigal_url)
        database = madrigalweb.MadrigalData(madrigal_url)
        exp_list = database.getExperiments(
            instrument_code,
            dt_fr.year, dt_fr.month, dt_fr.day, dt_fr.hour, dt_fr.minute, dt_fr.second,
            dt_to.year, dt_to.month, dt_to.day, dt_to.hour, dt_to.minute, dt_to.second,
            local=0
        )
    return exp_list, madrigal_url, database


"""
Functions for hdf5 files
"""


def show_hdf5_structure(filename, filepath=''):
    """
    Show madrigal hdf5 file structure in console.
    Example:
        fn = "/home/leicai/01_work/00_data/madrigal/DMSP/20151102/dms_20151102_16s1.001.hdf5"
        show_structure(fn)
    """
    with h5py.File(os.path.join(filepath, filename), 'r') as fh5:
        pybasic.dict_print_tree(fh5, value_repr=True, dict_repr=True, max_level=None)


def show_hdf5_metadata(filename, filepath='', fields=None):
    """
    Show madrigal hdf5 file metadata values.
    Example:
        fn = "/home/leicai/01_work/00_data/madrigal/DMSP/20151102/dms_20151102_16s1.001.hdf5"
        show_metadata(fn)
    """
    with h5py.File(os.path.join(filepath, filename), 'r') as fh5:
        if "metadata" in fh5.keys():
            metadata_key = "metadata"
        elif "Metadata" in fh5.keys():
            metadata_key = "Metadata"
        else:
            print("Cannot find the key either metadata or Metadata!")
            return KeyError

        keys_in = fh5[metadata_key].keys()
        if fields is None:
            keys = keys_in
        else:
            keys = fields
        print("\x1b[0;31;40m" + pybasic.retrieve_name(fh5) + "\x1b[0m")
        # for key in keys:
        #    print("\x1b[1;33;40m" + "Metadata<--" + key + ": " + "\x1b[0m")
        #    print(fh5[metadata_key][key][:])
        pybasic.dict_print_tree(fh5[metadata_key], full_value=True)


def show_hdf5_group(filename, filepath='', group_name=""):
    """
    Show madrigal hdf5 file group information.
    Example:
        fn = "/home/leicai/01_work/00_data/madrigal/DMSP/20151102/dms_20151102_16s1.001.hdf5"
        show_metadata(fn, group_name="Metadata")
    """
    with h5py.File(os.path.join(filepath, filename), 'r') as fh5:
        if group_name not in fh5.keys():
            print("Cannot find the key either metadata or Metadata!")
            return KeyError

        keys_in = fh5[group_name].keys()
        print("\x1b[0;31;40m" + pybasic.retrieve_name(fh5) + "\x1b[0m")
        # for key in keys:
        #    print("\x1b[1;33;40m" + "Metadata<--" + key + ": " + "\x1b[0m")
        #    print(fh5[metadata_key][key][:])
        pybasic.dict_print_tree(fh5[group_name], full_value=True)


if __name__ == "__main__":
    fn = "/Users/lcai/Downloads/EISCAT_2021-03-10_beata_ant@uhfa.hdf5"
    #show_hdf5_structure(fn)
    #show_hdf5_metadata(fn)
    #show_hdf5_group(fn, group_name="figures")
    list_all_instruments()





# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

from geospacelab.datahub.sources.esa_eo.swarm.loader import LoaderModel

# define the default variable name dictionary
default_variable_name_dict = {
    'CDF_EPOCH': 'Timestamp',
    'SC_GEO_LAT': 'Latitude',
    'SC_GEO_LON': 'Longitude',
    'SC_GEO_r': 'Radius',
    'SC_QD_LAT': 'QDLatitude',
    'SC_QD_MLT': 'MLT',
    'v_i_H_x': 'Vixh',
    'v_i_H_x_err': 'Vixh_error',
    'v_i_V_x': 'Vixv',
    'v_i_V_x_err': 'Vixv_error',
    'v_i_H_y': 'Viy',
    'v_i_H_y_err': 'Viy_error',
    'v_i_V_z': 'Viz',
    'v_i_V_z_err': 'Viz_error',
    'v_SC_N': 'VsatN',
    'v_SC_E': 'VsatE',
    'v_SC_C': 'VsatC',
    'E_H_x': 'Ehx',
    'E_H_y': 'Ehy',
    'E_H_z': 'Ehz',
    'E_V_x': 'Evx',
    'E_V_y': 'Evy',
    'E_V_z': 'Evz',
    'B_x': 'Bx',
    'B_y': 'By',
    'B_z': 'Bz',
    'v_i_CR_x': 'Vicrx',
    'v_i_CR_y': 'Vicry',
    'v_i_CR_z': 'Vicrz',
    'QUALITY_FLAG': 'Quality_flags',
    'CALIB_FLAG': 'Calibration_flags',
}


class Loader(LoaderModel):
    """
    Load SWARM 2Hz or 16HZ TII data products. Currently support versions higher than "0301".

    The class is a hierarchy of :class:`SWARM data LoaderModel <geospacelab.datahub.sources.esa_eo.swarm.loader.LoaderModel>`

    """
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('variable_name_dict', default_variable_name_dict)
        super(Loader, self).__init__(*args, **kwargs)
        self.variables['SC_GEO_r'] = self.variables['SC_GEO_r'] * 1e-3  # in km
        self.variables['SC_GEO_ALT'] = self.variables['SC_GEO_r'] - 6371.2




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import copy
import weakref
import re

import numpy as np
from typing import Dict

import geospacelab.toolbox.utilities.pyclass as pyclass
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pybasic as basic


class VisualAxis(object):
    """
    The attribute class appended to the Visual class for setting the attributes along an axis.


    """

    def __init__(self):
        self.data = None
        self.data_err = None
        self.data_scale = 1.
        self.data_res = None
        self.mask_gap = None
        self.label = ''
        self.label_style = 'double'  # or 'single
        self.label_pos = None  # label position
        self.unit = ''
        self.lim = None
        self.scale = 'linear'
        self.invert = False
        self.ticks = None
        self.tick_labels = None
        self.minor_ticks = None
        self.major_tick_max = 6
        self.minor_tick_max = None
        self.reverse = False
        self.visible = True

    def config(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)


class VisualPlotConfig(object):
    """
    The attribute class appended to the Visual class for setting the plots.


    """

    def __init__(self):
        self.visible = True
        self.line = {}
        self.pattern = {}
        self.errorbar = {}
        self.pcolormesh = {}
        self.imshow = {}
        self.scatter = {}
        self.legend = {}
        self.colorbar = {}
        self.bar = {}
        self.scatter = {}
        self.fill_between = {}
        self.style = None

    def config(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)


class NDim(int):
    def __new__(cls, value, extra):
        return float.__new__(cls, value)

    def __init__(self, value, extra):
        float.__init__(value)
        self.extra = extra


class Visual(object):
    """
    The Visual class is used for setting the visualization attributes appending to a variable.


    """

    def __init__(self, **kwargs):
        self._axis = {}
        self.variable = None
        self._ndim = None

        self.variable = kwargs.pop('variable', None)
        self.axis = kwargs.pop('axis', None)
        self.plot_config = kwargs.pop('plot_config', VisualPlotConfig())
        self.ndim = kwargs.pop('ndim', self._ndim)

    def config(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)

    def add_attr(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=True, logging=logging, **kwargs)

    def clone(self):
        kwargs = {
            'variable': self.variable,
            'axis': copy.deepcopy(self.axis),
            'plot_config': copy.deepcopy(self.plot_config),
            'ndim': copy.deepcopy(self.ndim)
        }

        return self.__class__(**kwargs)

    @property
    def variable(self):
        if self._variable_ref is None:
            return None
        else:
            return self._variable_ref()

    @variable.setter
    def variable(self, var_obj):
        if var_obj is None:
            self._variable_ref = None
            return
        if issubclass(var_obj.__class__, VariableBase):
            self._variable_ref = weakref.ref(var_obj)
        else:
            raise TypeError

    @property
    def ndim(self) -> int:
        _ndim = self._ndim
        if _ndim is None:
            if self.variable is not None:
                _ndim = self.variable.ndim
        return _ndim

    @ndim.setter
    def ndim(self, value):
        self._ndim = value

    @property
    def axis(self) -> Dict[int, VisualAxis]:
        if not dict(self._axis):
            ndim = self.ndim
            if ndim is None:
                try:
                    ndim = self.variable.ndim
                    if ndim is None:
                        return None
                    self.ndim = ndim
                except AttributeError:
                    return None
            if ndim < 2:
                ndim = 2
            for ind in range(ndim + 1):
                self._axis[ind] = VisualAxis()
        return self._axis

    @axis.setter
    def axis(self, a_dict):
        if a_dict is None:
            return
        if type(a_dict) is dict:
            for key, value in a_dict.items():
                if not type(key) is int:
                    raise TypeError
                if value is None:
                    self._axis[key] = VisualAxis()
                elif type(value) is dict:
                    self._axis[key].config(**value)
                elif isinstance(value, VisualAxis):
                    self._axis[key] = value
                else:
                    raise TypeError
        else:
            raise TypeError

    @property
    def plot_config(self):
        return self._plot

    @plot_config.setter
    def plot_config(self, value):
        if value is None:
            self._plot = VisualPlotConfig()
            return
        if isinstance(value, VisualPlotConfig):
            self._plot = value
        elif type(value) is dict:
            self._plot.config(**value)
        else:
            raise TypeError


class VariableBase(object):
    """VariableModel is a base class for a geospace variable with useful attributes

    :ivar str name: The variable's name.
    :ivar str fullname: The variable's full name.
    :ivar str  label: The variable's label for display. If a raw string (e.g., r'$\alpha$'),
        it will show a latex format.
    :ivar str data_type: The variable's data type in one of them: 'support_data', 'data', and 'metadata'
        in the same format as in a NASA's cdf file.
    :ivar str group: The group that the variable is belonged to, e.g., var.name = 'v_i_z', var.group = 'ion velocity',
        as the label in plots with multiple lines. ['']
    :ivar str  unit: The variable's unit. ['']
    :ivar str  unit_label: The unit's  label,  used for plots. The string is a raw string (e.g., r'$n_e$').
                If None, the plot will use unit as a label.
    :ivar Quantity object  quantity: The physical quantity associated with the variable, waiting for implementing. [None]
    :ivar np.ndarray  value: the variable's value.  The default type is a np.ndarray.
        The axis=0 along the time, axis=1 along height, lat,
        lon. For a scalar, value in a shape of (1, ).
    :ivar str or np.ndarray  error: the variable's error. Either a np.ndarray or a string. When it's a string, the string is a variable name
        indicating the variable in the associated dataset (see :attr:`dataset` below).
    :ivar int ndim: The number of dimension.
    :ivar dict depends: The full depends of the variables. Usually Axis 0 for time, next for spatial distributions,
        and then for components.
    :ivar Dataset object dataset: The dataset that the variable is appended.
    :ivar Visual object visual: the attributes for visualization.

    """

    __visual_model__ = Visual
    __dataset_model__ = None

    def __init__(
            self, value=None, error=None, data_type=None,
            name='', fullname='', label='', group='',
            unit=None, unit_label=None, quantity=None,
            variable_type='scalar',
            ndim=None, depends=None, dataset=None, visual=None,
            **kwargs):
        """Initial settings

            :param  name: The variable's name, ['']
            :type   name: str
            :param  fullname: The variable's full name, ['']
            :type   fullname: str
            :param  label: The variable's label for display. If a raw string (e.g., r'$\alpha$'),
                it will show a latex format.
            :type   label: str
            :param  data_type: The variable's data type in one of them: 'support_data', 'data', and 'metadata'
                in the same format as in a NASA's cdf file. ['']
            :type   data_type: str
            :param  group: The group that the variable is belonged to, e.g., var.name = 'v_i_z', var.group = 'ion velocity',
                as the label in plots with multiple lines. ['']
            :type   group: str
            :param  unit: The variable's unit. ['']
            :type   unit: str
            :param  unit_label: The unit's  label,  used for plots. The string is a raw string (e.g., r'$n_e$').
                        If None, the plot will use unit as a label.
            :type   unit_label: str
            :param  quantity: The physical quantity associated with the variable, waiting for implementing. [None]
            :type   quantity: TBD.
            :param  value: the variable's value. Usually it's a np.ndarray. The axis=0 along the time, axis=1 along height, lat,
                lon. For a scalar, value in a shape of (1, ).
            :type   value: np.ndarray, default: None
            :param  error: the variable's error. Either a np.ndarray or a string. When it's a string, the string is a variable name
                indicating the variable in the associated dataset (see :attr:`dataset` below).
            :type error: str or np.ndarry
            :param ndim: The number of dimension.
            :type ndim: int
            :param depends: The full depends of the variables. Usually Axis 0 for time, next for spatial distributions,
                and then for components.
            :type depends: dict
            :param dataset: The dataset that the variable is appended.
            :type dataset: DatasetModel object
            :param visual: the attributes for visualization.
            :type visual: dict or Visual object, default: None.
        """
        # set default values

        from geospacelab.datahub.__dataset_base__ import DatasetBase

        self.__dataset_model__ = DatasetBase

        self.name = name
        self.fullname = fullname

        self.label = label

        self.data_type = data_type  # 'support_data', 'data', 'metadata'
        self.group = group

        self.unit = unit
        self.unit_label = unit_label

        self.quantity = quantity

        self.value = value
        self.error = error

        self.variable_type = variable_type  # scalar, vi, tensor, ...
        self.ndim = ndim
        self._depends = {}
        self.depends = depends

        self.dataset = dataset

        self._visual = None
        self.visual = visual
        self._attrs_registered = ['name', 'fullname', 'label', 'data_type', 'group', 'unit', 'unit_label',
                                  'quantity', 'value', 'error', 'variable_type', 'ndim', 'depends', 'dataset',
                                  'visual']

    def config(self, logging=True, **kwargs):
        """
        Configure the variable attributes. If the attribute is not the default attributes, return error.

        :param logging: If True, show logging.
        :param kwargs: A dictionary of the attributes.
        :return: None
        """
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)

    def add_attr(self, logging=True, **kwargs):
        """
        Similar as config, but add a new attribute.
        """

        self._attrs_registered.extend(kwargs.keys())
        pyclass.set_object_attributes(self, append=True, logging=logging, **kwargs)

    def clone(self, omit_attrs=None):
        """
        Clone a variable and return a new instance.

        :param omit_attrs: The attributes omitted to be copied. If None, copy all.
        :return: new variable instance
        """
        if omit_attrs is None:
            omit_attrs = []
        kwargs = {}
        for key in self._attrs_registered:
            if key in omit_attrs:
                continue
            if key == 'visual':
                kwargs['visual'] = self.visual.clone()
            elif key == 'dataset':
                kwargs['dataset'] = self.dataset
            else:
                kwargs[key] = copy.deepcopy(getattr(self, key))
        return self.__class__(**kwargs)

    def get_depend(self, axis=None, retrieve_data=True):
        """
        Get the dependence of the variable along an axis.

        :param axis: The axis.
        :param retrieve_data: If True, Convert identifier to data.
        :return: A dictionary of dependence.
        """
        # axis = 0, 1, 2
        if axis in self.depends.keys():
            depend = self.depends[axis]
        else:
            return None
        depend_new = copy.deepcopy(depend)
        if retrieve_data:
            for key, value in depend.items():
                if isinstance(value, str):
                    try:
                        value = self.dataset[value].value
                    except KeyError:
                        print('The variable {} has not been assigned!'.format(key))
                        value = None
                depend_new[key] = copy.deepcopy(value)
        return depend_new

    def set_depend(self, axis, depend_dict):
        """
        Set the dependence along an axis.

        :param axis: The axis.
        :param depend_dict: the dictionary of the dependence, e.g., {'DATETIME': 'DATETIME'}.
        :return:
        """
        if not isinstance(depend_dict, dict):
            raise TypeError
        self.depends[axis] = depend_dict

    def _string_parse_for_attr(self, string):
        if not str(string):
            return string
        rc = re.compile(r'^@(.+)')
        rm = rc.match(string)
        if rm is None:
            return string

        results = rm.groups()[0].split('.')

        if results[0].lower() in ['v', 'var', 'variable']:
            attr = getattr(self, results[1])
        elif results[0].lower() in ['d', 'dataset']:
            if len(results) == 2:
                results.append('value')
            attr = getattr(self.dataset[results[1]], results[2])
        elif results[0].lower() in ['vd', 'depends']:
            depend = self.get_depend(axis=int(results[1]))
            attr = depend[results[2]]
        return attr
        #
        # if string[0] == '@':
        #     splits = string[1:].split('.')
        #     if splits[0] in ['v']:
        #         result = getattr(self, splits[1])
        #     elif splits[0] in ['d']:
        #         result = getattr(self.dataset[splits[1]], splits[2])
        #     else:
        #         raise ValueError
        # else:
        #     result = string
        return result

    def get_visual_axis_attr(self, attr_name=None, axis=None):
        """
        Get the visual attributes along an axis.

        :param attr_name:
        :param axis:
        :return:
        """
        attr = getattr(self.visual.axis[axis], attr_name)
        type_attr = type(attr)
        if type_attr is not list:
            attr = [attr]
        results = []
        for a in attr:
            if type(a) is str:
                result = self._string_parse_for_attr(a)
            else:
                result = a
            results.append(copy.deepcopy(result))
        if type_attr is not list:
            results = results[0]
        return results

    def join(self, var_new):
        """
        Join a numpy array or a variable instance to the current variable.
        :param var_new: The input variable.
        :type var_new: list, np.ndarray, or variable instance
        :return:
        """
        if issubclass(var_new.__class__, VariableBase):
            v = var_new.value
        else:
            v = var_new

        if self.value is None:
            self.value = v
            return

        if type(v) is np.ndarray:
            self.value = np.concatenate((self.value, v), axis=0)
        else:
            if np.isscalar(v):
                v = tuple([v])
            elif isinstance(v, (list, tuple)):
                v = tuple(v)
            if v != self.value:
                mylog.StreamLogger.warning("The scalar variables have different values!")
            return

    def flatten(self):
        return self.value.flatten()

    def __repr__(self):
        value_repr = repr(self.value)
        rep = f"GeospaceLab Variable object <name: {self.name}, value: {value_repr}, unit: {self.unit}>"
        return rep

    @property
    def visual(self) -> __visual_model__:
        return self._visual

    @visual.setter
    def visual(self, value):

        if value == 'new':
            self._visual = Visual(variable=self)
        elif value == 'on':
            if not isinstance(self._visual, Visual):
                self.visual = 'new'
        elif isinstance(value, Visual):
            self._visual = value
        elif type(value) is dict:
            if not isinstance(self._visual, Visual):
                self.visual = 'new'
            pyclass.set_object_attributes(self.visual, append=False, logging=True, **value)
        elif value == 'off' or value is None:
            self._visual = None
        else:
            raise ValueError

    @property
    def dataset(self):
        if self._dataset_ref is None:
            return None
        else:
            return self._dataset_ref()

    @dataset.setter
    def dataset(self, dataset_obj):
        if dataset_obj is None:
            self._dataset_ref = None
            return

        if issubclass(dataset_obj.__class__, self.__dataset_model__):
            self._dataset_ref = weakref.ref(dataset_obj)
        else:
            raise TypeError

    @property
    def value(self) -> np.ndarray:
        if self._value is None:
            return None
        elif isinstance(self._value, str):
            if self.dataset is not None:
                return self.dataset[self._value].value
            else:
                return self._value
        else:
            return self._value

    @value.setter
    def value(self, v):
        if v is None:
            self._value = None
            return
        # type check
        if type(v) is str:
            self._value = v
            return
        if not isinstance(v, np.ndarray):
            if basic.isnumeric(v):
                v = tuple([v])
            elif isinstance(v, (list, tuple)):
                v = tuple(v)
            else:
                raise TypeError
            self._value = v
            return
            # reshape np.array with shape like (m,) m>1
        # if len(v.shape) == 1 and v.shape != (1,):
        #     v = v.reshape((v.shape[0], 1))
        self._value = v

    @property
    def error(self) -> np.ndarray:
        if self._error is None:
            return None
        elif isinstance(self._error, str):
            if self.dataset is not None:
                return self.dataset[self._error].value
            else:
                return self._error
        else:
            return self._error

    @error.setter
    def error(self, v):
        if v is None:
            self._error = None
            return
        # type check
        if type(v) is str:
            self._error = v
            return
        if not isinstance(v, np.ndarray):
            if basic.isnumeric(v):
                v = np.array([v])
            elif isinstance(v, (list, tuple)):
                v = np.array(v)
            else:
                raise TypeError
        # reshape np.array with shape like (m,) m>1
        if len(v.shape) == 1 and v.shape != (1,):
            v = v.reshape((v.shape[0], 1))
        self._error = v

    @property
    def unit_label(self):
        label = self._unit_label
        if not str(label):
            label = self.unit
        return label

    @unit_label.setter
    def unit_label(self, label):
        if label is None:
            self._unit_label = ''
            return
        if type(label) is not str:
            raise TypeError
        self._unit_label = label

    @property
    def label(self):
        lb = self._label
        if not str(lb):
            lb = self.name
        return lb

    @label.setter
    def label(self, lb):
        if lb is None:
            self._label = ''
            return
        if type(lb) is not str:
            raise TypeError
        self._label = lb

    @property
    def variable_type(self):
        return self._variable_type

    @variable_type.setter
    def variable_type(self, value):
        if value is None:
            self._variable_type = None
            return
        if value.lower() not in ['scalar', 'vi', 'matrix', 'tensor']:
            raise AttributeError
        self._variable_type = value.lower()

    @property
    def ndim(self):
        if self._ndim is None:
            var_type = self.variable_type
            if var_type is None:
                mylog.StreamLogger.warning('the variable type has not been defined! Use default type "scalar"...')
                var_type = "scalar"
            value = self.value
            if value is None:
                mylog.StreamLogger.warning("Variable's value was not assigned. Return None!")
                return None
            shape = value.shape
            offset = 0
            if var_type == 'scalar':
                offset = -1
            elif var_type == 'vi':
                offset = -1
            elif var_type == 'matrix':
                offset = -2
            if shape == (1,):
                nd = 0
            else:
                nd = len(shape) + offset
            if nd <= 0:
                raise ValueError("ndim cannot be a non-positive integer")
            self._ndim = nd
        return self._ndim

    @ndim.setter
    def ndim(self, value):
        if value is None:
            self._ndim = None
            return
        if type(value) is not int:
            raise TypeError
        self._ndim = value

    @property
    def depends(self):
        return self._depends

    @depends.setter
    def depends(self, d_dict):
        if d_dict is None:
            self._depends = {}
            return

        if type(d_dict) is not dict:
            raise TypeError

        for key, value in d_dict.items():
            if type(key) is not int:
                raise KeyError
            if value is None:
                self._depends[key] = {}
            else:
                self._depends[key] = value


#
#
# class VariableBase(np.ndarray):
#     _attrs_registered = [
#         'name', 'fullname', 'label',
#         'data_type', 'group',
#         'unit', 'unit_label',
#         'quantity',
#         'value', 'error',
#         'variable_type',
#         'kind',
#         'depends', 'dataset',
#         'visual'
#     ]
#
#     def __new__(cls, arr, copy=True, dtype=None, order='C', subok=False, ndmin=0, **kwargs):
#         if issubclass(arr.__class__, VariableBase):
#             obj_out = arr.clone()
#         else:
#             obj_out = np.array(arr, copy=copy, dtype=dtype, order=order, subok=subok, ndmin=ndmin)
#             obj_out = obj_out.view(cls)
#
#         obj_out.config(**kwargs)
#
#         return obj_out
#
#     def config(self, logging=True, **kwargs):
#         """
#         Configure the variable attributes. If the attribute is not the default attributes, return error.
#
#         :param logging: If True, show logging.
#         :param kwargs: A dictionary of the attributes.
#         :return: None
#         """
#
#         # check registered attributes
#         for key, value in kwargs.items():
#             if key not in self._attrs_registered:
#                 raise KeyError(f'The attribute {key} is not registered! Use "add_attr()" to add a new attribute.')
#
#         pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)
#
#     def add_attr(self, logging=True, **kwargs):
#         """
#         Similar as config, but add a new attribute.
#         """
#
#         self._attrs_registered.extend(kwargs.keys())
#         pyclass.set_object_attributes(self, append=True, logging=logging, **kwargs)
#
#     def copy_attr(self, obj, to_whom=None, omit_attrs=None):
#         """
#         Copy the attributes from obj to the current instance.
#
#         :param obj: the VariableBase or its subclass instance
#         :type obj: VariableBase instance
#         :param to_whom: The destination instance
#         :type to_whom: VariableBase instance. If ``None``, copy to self
#         :param omit_attrs: The attributes omitted. If None, copy all
#         :type omit_attrs: dict or None
#         """
#
#         if omit_attrs is None:
#             omit_attrs = {}
#
#         if to_whom is None:
#             to_whom = self
#
#         for key in self._attrs_registered:
#             if key in omit_attrs.keys():
#                 continue
#             if key == 'visual':
#                 to_whom.visual = obj.visual.clone()
#             elif key == 'dataset':
#                 to_whom.dataset = obj.dataset
#             else:
#                 setattr(to_whom, key, copy.deepcopy(getattr(obj, key)))
#
#         return to_whom
#
#     def clone(self, omit_attrs=None):
#         """
#         Clone a variable and return a new instance.
#
#         :param omit_attrs: The attributes omitted to be copied. If None, copy all.
#         :return: new variable instance
#         """
#
#         obj_new = self.__class__(self.view(np.ndarray))
#         obj_new.copy_attr(self, omit_attrs=omit_attrs)
#
#         return obj_new
#
#     def __array_ufunc__(self, ufunc, method, *inputs, out=None, **kwargs):
#         """
#         This implementation of __array_ufunc__ makes sure that all custom attributes are maintained when a ufunc operation is performed on our class.'''
#         See also :ref:`Propagate attributes <https://stackoverflow.com/questions/51520630/subclassing-numpy-array-propagate-attributes>`.
#         """
#
#         # convert inputs and outputs of class ArraySubclass to np.ndarray to prevent infinite recursion
#         args = ((i.view(np.ndarray) if issubclass(i, VariableBase) else i) for i in inputs)
#         outputs = kwargs.pop('out', None)
#         if outputs:
#             kwargs['out'] = tuple((o.view(np.ndarray) if issubclass(o, VariableBase) else o) for o in outputs)
#         else:
#             outputs = (None,) * ufunc.nout
#
#         # call numpys implementation of __array_ufunc__
#         results = super().__array_ufunc__(ufunc, method, *args, **kwargs)  # pylint: disable=no-member
#         if results is NotImplemented:
#             return NotImplemented
#         if method == 'at':
#             # method == 'at' means that the operation is performed in-place. Therefore, we are done.
#             return
#         # now we need to make sure that outputs that where specified with the 'out' argument are handled corectly:
#         if ufunc.nout == 1:
#             results = (results,)
#         results = tuple((self._copy_attrs_to(result) if output is None else output)
#                         for result, output in zip(results, outputs))
#         return results[0] if len(results) == 1 else results
#
#     def __array_finalize__(self, obj):
#         """
#         Used for handling new instances created by three ways:
#         * Explicit constructor:
#             * self type is ´´cls´´
#             * obj type is ``None``
#         * View casting:
#             * self type is ``cls``
#             * obj type is ``np.ndarray``
#         * Slicing (new from template)
#             * self type is ``cls``
#             * obj type is ``cls``
#         """
#
#         # Nothing needed for new variable or np.ndarray.view()
#         if obj is None or isinstance(obj.__class__, np.ndarray):
#             return None
#
#         if issubclass(obj.__class__, self.__class__):
#             self.copy_attr(obj)
#         # else:
#         #     for attr in getattr(self.__class__, '_attrs_registered'):
#         #         self.__setattr__(attr, cp.deepcopy(getattr(self.__class__, attr))
#
#
#
#
# class Visual_1(object):
#     def __init__(self, **kwargs):
#         self.plot_type = None
#         self.x_data = None
#         self.y_data = None
#         self.z_data = None
#         self.x_data_scale = 1
#         self.y_data_scale = 1
#         self.z_data_scale = 1
#         self.x_data_res = None
#         self.y_data_res = None
#         self.z_data_res = None
#         self.x_err_data = None
#         self.y_err_data = None
#         self.z_err_data = None
#         self.x_lim = None
#         self.y_lim = None
#         self.z_lim = None
#         self.x_label = None
#         self.y_label = None
#         self.z_label = None
#         self.x_scale = None
#         self.y_scale = None
#         self.z_scale = None
#         self.x_unit = None
#         self.y_unit = None
#         self.z_unit = None
#         self.x_ticks = None
#         self.y_ticks = None
#         self.z_ticks = None
#         self.x_tick_labels = None
#         self.y_tick_labels = None
#         self.z_tick_labels = None
#         self.color = None
#         self.visible = True
#         self.plot_config = {}
#         self.axis_config = {}
#         self.legend_config = {}
#         self.colorbar_config = {}
#         self.errorbar_config = {}
#         self.pcolormesh_config = {}
#         self.config(**kwargs)
#
#     def config(self, logging=True, **kwargs):
#         pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)
#
#     def add_attr(self, logging=True, **kwargs):
#         pyclass.set_object_attributes(self, append=True, logging=logging, **kwargs)

#
# def to_dict(self):
#     class_vars = vars(Visual)  # get any "default" attrs defined at the class level
#     inst_vars = vars(self)  # get any attrs defined on the instance (self)
#     all_vars = dict(class_vars)
#     all_vars.update(inst_vars)
#     # filter out private attributes
#     public_vars = {k: v for k, v in all_vars.items() if not k.startswith('_')}
#     del public_vars['set_attr']
#     del public_vars['to_dict']
#     return public_vars


#
#
# class Variable(np.ndarray, npmixin.NDArrayOperatorsMixin):
#     _attrs_registered = [
#         'dataset',
#         'name', 'label', 'description', 'group',
#         'error',
#         'depends', 'unit', 'quantity_type',
#         'cs',
#         'visual'
#     ]
#
#     _dataset = None
#     _name = ''
#     _label = ''
#     _description = ''
#     _group = ''
#     _error = None
#     _unit = ''
#     _quantity_type = None
#     _cs = None
#     _depends = []
#     _visual = None
#
#     def __new__(cls, arr, copy=True, dtype=None, order='C', subok=False, ndmin=0, **kwargs):
#
#         if issubclass(arr.__class__, Variable):
#             obj_out = arr
#         else:
#             obj_out = np.array(arr, copy=copy, dtype=dtype, order=order, subok=subok, ndmin=ndmin)
#             obj_out = obj_out.view(cls)
#
#         obj_out.set_attr(**kwargs)
#         return obj_out
#
#     def __array_finalize__(self, obj):
#         # nothing needed for new variable or np.ndarray.view()
#         if obj is None or isinstance(obj.__class__, np.ndarray):
#             return None
#
#         if issubclass(obj.__class__, Variable):
#             for attr in getattr(obj, '_attrs_registered'):
#                 self.__setattr__(attr, cp.deepcopy(getattr(obj, attr)))
#         # else:
#         #     for attr in getattr(self.__class__, '_attrs_registered'):
#         #         self.__setattr__(attr, cp.deepcopy(getattr(self.__class__, attr)))
#
#     def __array_ufunc__(self, ufunc, method, *inputs, out=None, **kwargs):
#
#         args = []
#         in_no = []
#         attr_config = kwargs.pop('attr_config', None)
#
#         for i, input_ in enumerate(inputs):
#             if issubclass(input_.__class__, Variable):
#                 in_no.append(i)
#                 args.append(input_.view(np.ndarray))
#             else:
#                 args.append(input_)
#
#         outputs = out
#         out_no = []
#
#         if outputs:
#             out_args = []
#             for j, output in enumerate(outputs):
#                 if issubclass(output.__class__, Variable):
#                     out_no.append(j)
#                     out_args.append(output.view(np.ndarray))
#                 else:
#                     out_args.append(output)
#             kwargs['out'] = tuple(out_args)
#         else:
#             outputs = (None,) * ufunc.nout
#
#         results = super(self.__class__, self).__array_ufunc__(
#             ufunc, method, *args, **kwargs
#         )
#
#         if results is NotImplemented:
#             return NotImplemented
#
#         if ufunc.nout == 1:
#             results = (results,)
#
#         results = tuple((np.asarray(result).view(self.__class__)
#                          if output is None else output)
#                         for result, output in zip(results, outputs))
#
#         if len(results) == 1:
#             result = results[0]
#             result.copy_attr(self)
#             if bool(attr_config):
#                 result.set_attr(**attr_config)
#             return result
#         else:
#             return results
#
#     def __reduce__(self):
#         # patch to pickle Quantity objects (ndarray subclasses), see
#         # http://www.mail-archive.com/numpy-discussion@scipy.org/msg02446.html
#
#         object_state = list(super().__reduce__())
#         object_state[2] = (object_state[2], self.__dict__)
#         return tuple(object_state)
#
#     def __setstate__(self, state):
#         # patch to unpickle Variable objects (ndarray subclasses), see
#         # http://www.mail-archive.com/numpy-discussion@scipy.org/msg02446.html
#
#         nd_state, own_state = state
#         super().__setstate__(nd_state)
#         self.__dict__.update(own_state)
#
#     def __repr__(self):
#         str_pref = '<Geospace ' + self.__class__.__name__ + ' '
#         str_arr = np.array2string(self.view(np.ndarray), separator=', ',
#                                   prefix=str_pref)
#         return f'{str_pref}{str_arr}{self.unit:s}>'
#
#     def config(self, **kwargs):
#         self.set_attr(**kwargs)
#
#     def set_attr(self, **kwargs):
#         # set values for the registered attributes
#         attr_add = kwargs.pop('attr_add', True)
#         logging = kwargs.pop('logging', False)
#         new_attrs = []
#         for key, value in kwargs.items():
#             if key in self._attrs_registered:
#                 setattr(self, key, value)
#             elif attr_add:
#                 setattr(self, key, value)
#                 self._attrs_registered.append(key)
#                 new_attrs.append(key)
#
#         if logging:
#             if attr_add:
#                 logger.StreamLogger.info('Attrs: {} added!'.format(', '.join(new_attrs)))
#             else:
#                 logger.StreamLogger.warning('Attrs: {} not added!'.format(', '.join(new_attrs)))
#
#     def copy_attr(self, obj):
#         for attr in getattr(obj, '_attrs_registered'):
#             self.__setattr__(attr, cp.deepcopy(getattr(obj, attr)))
#
#     @property
#     def dataset(self):
#         return self._dataset
#
#     @dataset.setter
#     def dataset(self, obj):
#         from geospacelab.datahub import Dataset
#         # Check the type:
#         if obj is None:
#             self._dataset = None
#         elif not isinstance(obj, Dataset):
#             raise ValueError
#         self._dataset = obj
#
#     @property
#     def name(self):
#         return self._name
#
#     @name.setter
#     def name(self, str_):
#         # Check the type:
#         if str_ is None:
#             str_ = ''
#         elif not isinstance(str_, str):
#             raise ValueError
#         self._name = str_
#
#     @property
#     def label(self):
#         return self._label
#
#     @label.setter
#     def label(self, str_):
#         # Check the type:
#         if str_ is None:
#             str_ = ''
#         elif not isinstance(str_, str):
#             raise ValueError
#         self._label = str_
#
#     @property
#     def description(self):
#         return self._description
#
#     @description.setter
#     def description(self, str_):
#         # Check the type:
#         if str_ is None:
#             str_ = ''
#         elif not isinstance(str_, str):
#             raise ValueError
#         self._description = str_
#
#     @property
#     def group(self):
#         return self._group
#
#     @group.setter
#     def group(self, str_):
#         # Check the type:
#         if str_ is None:
#             str_ = ''
#         elif not isinstance(str_, str):
#             raise ValueError
#         self._group = str_
#
#     @property
#     def error(self):
#         return self._error
#
#     @error.setter
#     def error(self, err):
#         if err is None:
#             self._error = None
#         if isinstance(err, self.__class__):
#             self._error = err
#         elif isinstance(err, str):
#             self._error = self.dataset[err]
#
#     @property
#     def unit(self):
#         return self._unit
#
#     @unit.setter
#     def unit(self, value):
#         # Check the type:
#         if value is None:
#             value = ''
#         elif not isinstance(value, str):
#             raise ValueError
#         self._unit = value
#
#     @property
#     def quantity_type(self):
#         return self._quantity_type
#
#     @quantity_type.setter
#     def quantity_type(self, str_):
#         # Check the type:
#         if str_ is None:
#             str_ = ''
#         elif not isinstance(str_, str):
#             raise ValueError
#         self._quantity_type = str_
#
#     @property
#     def cs(self):
#         return self._cs
#
#     @cs.setter
#     def cs(self, obj):
#         # Check the type:
#         self._cs = obj
#
#     @property
#     def depends(self):
#         return self._depends
#
#     @depends.setter
#     def depends(self, value):
#         # Check the type:
#         if value is None:
#             value = []
#         elif not isinstance(value, list):
#             raise ValueError
#         self._depends = value
#
#     @property
#     def visual(self):
#         return self._visual
#
#     @visual.setter
#     def visual(self, obj):
#         # Check the type:
#         if obj is None:
#             self._visual = None
#         elif not isinstance(obj, Visual):
#             raise ValueError
#         self._visual = obj
#
#
# if __name__ == "__main__":
#     a = Variable([1, 2, 3, 4], name='a')
#     a = a.reshape((a.shape[0], 1))
#     b = 5 + a
#     c = np.sum(a)
#
#     pass



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import copy
from geospacelab.datahub import VariableModel as Var
import geospacelab.visualization.mpl.colormaps as cm

database = 'Madrigal'

timestamps = {
    'DATETIME': 'DATETIME',
}

coords = {
    'GEO': ['GEO_LAT', 'GEO_LON', 'GEO_ALT', 'GEO_ST'],
    'AACGM': ['AACGM_LAT', 'AACGM_LON', 'AACGM_R', 'AACGM_MLT'],
    'APEX': ['APEX_LAT', 'APEX_LON', 'APEX_ALT', 'APEX_MLT']
}

depend_0 = {'UT': 'DATETIME', 'TIME_1': 'DATETIME_1', 'TIME_2': 'DATETIME_2'}
depend_1 = {'HEIGHT': 'HEIGHT', 'RANGE': 'RANGE', 'GEO_LAT': 'GEO_LAT', 'GEO_LON': 'GEO_LON',
            'GEO_ALT': 'GEO_ALT', 'AACGM_LAT': 'AACGM_LAT', 'AACGM_LON': 'AACGM_LON'}

default_colormap = cm.cmap_jet_modified()

default_axis_dict_2d = {
    1:     {
        'data':     '@d.HEIGHT.value',
        'lim':      [90, 350],
        'label':    'h',
        'unit':     'km',
    },
    2:  {
        'data':     '@v.value',
        'label':    '@v.label',
        'unit':     '@v.unit',
    }
}

default_plot_config = {
    'line':         {
        'linestyle':        '',
        'linewidth':        1.5,
        'marker':           '.',
        'markersize':       3,
    },
    'pcolormesh':   {
        'cmap':            default_colormap,
    }
}

configured_variables = {}
visual = 'on'

####################################################################################################################
var_name = 'n_e'
var = Var(ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'electron density',
    'label': r'$n_e$',
    'unit': 'm-3',
    'unit_label': r'm$^{-3}$',
    'error': var_name + '_err',
    'depends': {0: depend_0, 1: depend_1},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_2d[1])
var.visual.axis[2].config(**default_axis_dict_2d[2])
var.visual.axis[2].scale = 'log'
var.visual.axis[2].lim = [8e9, 9e11]
# axis = var.visual.axis
# axis[1].data = "@d.height.value"
# axis[2].data = "@v.value"
# axis[1].lim = [90, 350]
# axis[1].label = 'h'
# axis[1].unit = 'km'
# axis[2].lim = [8e9, 9e11]
# axis[2].scale = 'log'
# axis[2].label = '@v.label'
# axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'T_i'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'ion temperature'
var.label = r'$T_i$'
var.unit = 'K'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.HEIGHT.value"
axis[2].data = "@v.value"
axis[1].lim = [90, 350]
axis[1].label = 'h'
axis[1].unit = 'km'
axis[2].lim = [100, 2500]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'T_e'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'electron temperature'
var.label = r'$T_e$'
var.unit = 'K'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.HEIGHT.value"
axis[2].data = "@v.value"
axis[1].lim = [90, 350]
axis[1].label = 'h'
axis[1].unit = 'km'
axis[2].lim = [100, 3500]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_los'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Line-of-sight ion velocity'
var.label = r'$v_i^{los}$'
var.unit = 'm/s'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.HEIGHT.value"
axis[2].data = "@v.value"
axis[1].lim = [90, 350]
axis[1].label = 'h'
axis[1].unit = 'km'
axis[2].lim = [-400, 400]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'AZ'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'azimuthal angle'
var.label = 'AZ'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-5, 365]
axis[1].ticks = [0, 90, 180, 270, 360]
axis[1].label = '@v.group'
axis[1].unit = ''
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'EL'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'elevation angle'
var.label = 'EL'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-5, 185]
axis[1].ticks = [0, 45, 90, 135, 180]
axis[1].label = '@v.group'
axis[1].unit = ''
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'P_Tx'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Transmitter Power'
var.label = 'Power'
var.unit = 'kW'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
plot_config.line['linewidth'] = 2.5
plot_config.line['linestyle'] = '-'
plot_config.line['marker'] = ''
plot_config.line['alpha'] = 0.5
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [0, 1600]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################

var_name = 'T_SYS_1'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'System temperature 1'
var.label = r'$T_{1}^{sys}$'
var.unit = 'K'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.line['linewidth'] = 2.5
plot_config.line['linestyle'] = '-'
plot_config.line['marker'] = ''
plot_config.line['alpha'] = 0.5

plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [0, 500]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import netCDF4 as nc
import datetime
import numpy as np
import pathlib
import re
import cftime
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog


class Loader(object):

    def __init__(self, file_path, file_ext='nc', pole='N', append_support_data=True):

        self.variables = {}
        self.metadata = {}
        self.file_path = pathlib.Path(file_path)
        self.file_ext = file_ext
        self.pole = pole
        self.append_support_data = append_support_data

        self.load_data()

    def load_data(self):
        if self.file_path.suffix.strip('.') in ['dat', 'txt'] or self.file_ext == 'ascii':
            file_nc = pathlib.Path(self.file_path.with_suffix('.nc'))
            if not file_nc.is_file():
                self.save_to_nc(self.file_path)
            self.file_ext = 'nc'
            self.file_path = file_nc
            
        if self.file_ext == 'nc':
            self.load_data_nc()
    
    def load_data_nc(self):

        dataset = nc.Dataset(self.file_path)
        variables = {}

        unix_time = dataset.variables['UNIX_TIME'][::]
        dts = dttool.convert_unix_time_to_datetime_cftime(unix_time)
        ntime = dts.size
        variables['DATETIME'] = np.reshape(dts, (ntime, 1))

        variables['GRID_MLAT'] = np.array(dataset.variables['MLAT'][::])
        variables['GRID_MLON'] = np.array(dataset.variables['MLON'][::])
        variables['GRID_MLT'] = np.array(dataset.variables['MLT'][::])
        variables['GRID_E_E'] = np.array(dataset.variables['E_E'][::])
        variables['GRID_E_E'] = np.array(dataset.variables['E_E'][::])
        variables['GRID_E_N'] = np.array(dataset.variables['E_N'][::])
        variables['GRID_v_i_E'] = np.array(dataset.variables['v_i_E'][::])
        variables['GRID_v_i_N'] = np.array(dataset.variables['v_i_E'][::])
        variables['GRID_phi'] = np.array(dataset.variables['phi'][::])

        if self.append_support_data:
            var_names_append = [
                'VCNUM', 'IMF_MODEL', 'CLOCK_ANGLE', 'DIP_TILT', 'E_SW',
                'SD_MODEL', 'FIT_ORDER', 'B_x_OMNI', 'B_y_OMNI', 'B_z_OMNI',
                'phi_CPCP', 'phi_MAX', 'phi_MIN'
            ]
            for var_name in var_names_append:
                variables[var_name] = np.array(dataset.variables[var_name][::]).reshape((ntime, 1))


        dataset.close()

        self.variables = variables

    def save_to_nc(self, file_path):
        if not file_path.is_file():
            raise FileExistsError
        with open(file_path, 'r') as f:
            text = f.read()

            # results = re.findall(
            #     r'^\s*(\d+)\s*\[(\d+),(\d+)]\s*([-\d.]+)\s*' +
            #     r'([-\d.]+)\s*([-\d.]+)\s*([-\d.]+)\s*([-\d.]+)\s*([-\d.]+)\s*([-\d.]+)\s*' +
            #     r'([\S]+)',
            #     text,
            #     re.M
            # )
            results = re.findall(
                r'^\s*(\d+)\s*\[(\d+),(\d+)]\s*([\S]+)\s*([\S]+)\s*([\S]+)\s*([\S]+)\s*([\S]+)\s*([\S]+)\s*([\S]+)\s*([\S]+)',
                text,
                re.M
            )
            results = list(zip(*results))
            nlat = 40
            nlon = 180
            ntime = len(results[0]) / nlon / nlat
            if ntime != int(ntime):
                raise ValueError
            ntime = int(ntime)
            mlat_arr = np.array(results[3]).reshape([ntime, nlat, nlon], order='C').transpose((0, 2, 1)).astype(np.float32)
            mlon_arr = np.array(results[4]).reshape([ntime, nlat, nlon], order='C').transpose((0, 2, 1)).astype(np.float32)
            EF_N_arr = np.array(results[5]).reshape([ntime, nlat, nlon], order='C').transpose((0, 2, 1)).astype(np.float32)
            EF_E_arr = np.array(results[6]).reshape([ntime, nlat, nlon], order='C').transpose((0, 2, 1)).astype(np.float32)
            v_N_arr = np.array(results[7]).reshape([ntime, nlat, nlon], order='C').transpose((0, 2, 1)).astype(np.float32)
            v_E_arr = np.array(results[8]).reshape([ntime, nlat, nlon], order='C').transpose((0, 2, 1)).astype(np.float32)
            phi_arr = np.array(results[9]).reshape([ntime, nlat, nlon], order='C').transpose((0, 2, 1)).astype(np.float32)

            dts = np.array(results[10])[::nlon * nlat]
            dts = [datetime.datetime.strptime(dtstr, "%Y-%m-%d/%H:%M:%S") for dtstr in dts]
            time_array = np.array(cftime.date2num(dts, units='seconds since 1970-01-01 00:00:00.0'))

            import aacgmv2
            mlt_arr = np.empty_like(mlat_arr)
            for i in range(ntime):
                mlt1 = aacgmv2.convert_mlt(mlon_arr[i].flatten(), dts[i]).reshape((nlon, nlat))
                mlt_arr[i, ::] = mlt1[::]

            fp = pathlib.Path(file_path.with_suffix('.nc'))
            fp.parent.resolve().mkdir(parents=True, exist_ok=True)
            fnc = nc.Dataset(fp, 'w')
            fnc.createDimension('UNIX_TIME', ntime)
            fnc.createDimension('MLAT', nlat)
            fnc.createDimension('MLON', nlon)

            fnc.title = "SuperDARN Potential maps"

            time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
            time.units = 'seconds since 1970-01-01 00:00:00.0'
            time[::] = time_array[::]

            mlat = fnc.createVariable('MLAT', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            mlat[::] = mlat_arr[::]
            mlon = fnc.createVariable('MLON', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            mlon[::] = mlon_arr[::]
            mlt = fnc.createVariable('MLT', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            mlt[::] = mlt_arr[::]
            EF_N = fnc.createVariable('E_N', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            EF_N[::] = EF_N_arr[::]
            EF_E = fnc.createVariable('E_E', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            EF_E[::] = EF_E_arr[::]
            v_N = fnc.createVariable('v_i_N', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            v_N[::] = v_N_arr[::]
            v_E = fnc.createVariable('v_i_E', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            v_E[::] = v_E_arr[::]
            phi = fnc.createVariable('phi', np.float32, ('UNIX_TIME', 'MLON', 'MLAT'))
            phi[::] = phi_arr[::]

            # Support data such as IMF, VCNUM, potential drop
            results = re.findall(
                r'^>[A-Za-z ]*\(VCNUM\):\s*([\d]+)',
                text,
                re.M
            )
            vcnum_arr = np.array(results, dtype=np.int32).reshape([ntime, 1], order='C')
            vcnum = fnc.createVariable('VCNUM', 'i4', ('UNIX_TIME',))
            vcnum[::] = vcnum_arr[::]

            results = re.findall(
                r'^>\s*IMF Model: ([\S]+) Bang\s*([\S]+) deg\., '
                + r'Esw\s*([\S]+) mV/m, tilt\s*([\S]+) deg\.,\s*([A-Za-z0-9]+), Fit Order:\s*([\d]+)',
                text,
                re.M
            )
            results = list(zip(*results))
            imf_model_arr = np.array(results[0], dtype=object).reshape([ntime, 1], order='C')
            imf_model = fnc.createVariable('IMF_MODEL', 'S8', ('UNIX_TIME',))
            imf_model[::] = imf_model_arr[::]
            B_angle_arr = np.array(results[1], dtype=np.float32).reshape([ntime, 1], order='C')
            B_angle = fnc.createVariable('CLOCK_ANGLE', np.float32, ('UNIX_TIME',))
            B_angle[::] = B_angle_arr[::]
            E_sw_arr = np.array(results[2], dtype=np.float32).reshape([ntime, 1], order='C')
            E_sw = fnc.createVariable('E_SW', np.float32, ('UNIX_TIME',))
            E_sw[::] = E_sw_arr[::]
            dipole_tilt_arr = np.array(results[3], dtype=np.float32).reshape([ntime, 1], order='C')
            dipole_tilt = fnc.createVariable('DIP_TILT', np.float32, ('UNIX_TIME',))
            dipole_tilt[::] = dipole_tilt_arr[::]
            SD_model_arr = np.array(results[4], dtype=object).reshape([ntime, 1], order='C')
            SD_model = fnc.createVariable('SD_MODEL', 'S8', ('UNIX_TIME',))
            SD_model[::] = SD_model_arr[::]
            fit_order_arr = np.array(results[5], dtype=np.int32).reshape([ntime, 1], order='C')
            fit_order = fnc.createVariable('FIT_ORDER', 'i4', ('UNIX_TIME',))
            fit_order[::] = fit_order_arr[::]

            results = re.findall(
                r'^> OMNI IMF:\s*Bx=([\S]+) nT,\s*By=([\S]+) nT,\s*Bz=([\S]+) nT',
                text,
                re.M
            )
            results = list(zip(*results))
            B_x_OMNI_arr = np.array(results[0], dtype=np.float32).reshape([ntime, 1], order='C')
            B_x_OMNI = fnc.createVariable('B_x_OMNI', np.float32, ('UNIX_TIME',))
            B_x_OMNI[::] = B_x_OMNI_arr[::]
            B_y_OMNI_arr = np.array(results[1], dtype=np.float32).reshape([ntime, 1], order='C')
            B_y_OMNI = fnc.createVariable('B_y_OMNI', np.float32, ('UNIX_TIME',))
            B_y_OMNI[::] = B_y_OMNI_arr[::]
            B_z_OMNI_arr = np.array(results[2], dtype=np.float32).reshape([ntime, 1], order='C')
            B_z_OMNI = fnc.createVariable('B_z_OMNI', np.float32, ('UNIX_TIME',))
            B_z_OMNI[::] = B_z_OMNI_arr[::]

            # > Potential: Drop = 33 kV, Min = -19 kV, Max = 14 kV
            results = re.findall(
                r'^> Potential:\s*Drop=([\S]+) kV,\s*Min=([\S]+) kV,\s*Max=([\S]+) kV',
                text,
                re.M
            )
            results = list(zip(*results))
            phi_CPCP_arr = np.array(results[0], dtype=np.float32).reshape([ntime, 1], order='C')
            phi_CPCP = fnc.createVariable('phi_CPCP', np.float32, ('UNIX_TIME',))
            phi_CPCP[::] = phi_CPCP_arr[::]
            phi_max_arr = np.array(results[1], dtype=np.float32).reshape([ntime, 1], order='C')
            phi_max = fnc.createVariable('phi_MAX', np.float32, ('UNIX_TIME',))
            phi_max[::] = phi_max_arr[::]
            phi_min_arr = np.array(results[2], dtype=np.float32).reshape([ntime, 1], order='C')
            phi_min = fnc.createVariable('phi_MIN', np.float32, ('UNIX_TIME',))
            phi_min[::] = phi_min_arr[::]

            print('From {} to {}.'.format(
                datetime.datetime.utcfromtimestamp(time_array[0]),
                datetime.datetime.utcfromtimestamp(time_array[-1]))
            )
            mylog.StreamLogger.info(
                "The requested SuperDARN map potential data has been saved in the file {}.".format(fp))
            fnc.close()

if __name__ == "__main__":
    import pathlib
    fp = pathlib.Path('/home/lei/afys-data/SuperDARN/PotentialMap/2016/SuperDARN_POTMAP_2min_20160315_N.nc')
    loader = Loader(file_path=fp)


    # if hasattr(readObj, 'pole'):
    #    readObj.filter_data_pole(boundinglat = 25)


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import pathlib
import numpy as np
import cdflib

import geospacelab.toolbox.utilities.pybasic as pybasic

default_variable_name_dict = {
}


class LoaderModel(object):
    """
    :param file_path: the full path of the data file
    :type file_path: pathlib.Path or str
    :param file_type: the type of the file, [cdf]
    :param variable_name_dict: the dictionary for mapping the variable names from the cdf files to the dataset
    :type variable_name_dict: dict
    :param direct_load: call the method :meth:`~.LoadModel.load_data` directly or not
    :type direct_load: bool
    """
    def __init__(self, file_path, file_type='cdf', variable_name_dict=None, direct_load=True, **kwargs):

        self.file_path = pathlib.Path(file_path)
        self.file_type = file_type
        self.variables = {}

        if variable_name_dict is None:
            variable_name_dict = default_variable_name_dict
        self.variable_name_dict = variable_name_dict

        if direct_load:
            self.load_data()

    def load_data(self, **kwargs):
        if self.file_type == 'cdf':
            self.load_cdf_data()

    def load_cdf_data(self):
        """
        load the data from the cdf file
        :return:
        """
        cdf_file = cdflib.CDF(self.file_path)
        cdf_info = cdf_file.cdf_info()
        variables = cdf_info['zVariables']

        if dict(self.variable_name_dict):
            new_dict = {vn: vn for vn in variables}
            self.variable_name_dict = pybasic.dict_set_default(self.variable_name_dict, **new_dict)

        for var_name, cdf_var_name in self.variable_name_dict.items():
            if var_name == 'CDF_EPOCH':
                epochs = cdf_file.varget(variable=cdf_var_name)
                epochs = cdflib.cdfepoch.unixtime(epochs)
                dts = [datetime.timedelta(seconds=epoch) + datetime.datetime(1970, 1, 1, 0, 0, 0) for epoch in epochs]
                self.variables['SC_DATETIME'] = np.array(dts).reshape((len(dts), 1))
                continue
            var = cdf_file.varget(variable=cdf_var_name)
            var = np.array(var)
            vshape = var.shape
            if len(vshape) == 1:
                var = var.reshape(vshape[0], 1)
            self.variables[var_name] = var





# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import tkinter as tk
from tkinter import ttk
from tkinter import filedialog as fd
from tkinter.messagebox import showinfo

# create the root window
root = tk.Tk()
root.title('Tkinter Open File Dialog')
root.resizable(False, False)
root.geometry('300x150')


def select_file(**kwargs):
    filetypes = (
        ('text files', '*.txt'),
        ('All files', '*.*')
    )

    filename = fd.askopenfilename(
        title='Open a file',
        initialdir='/',
        filetypes=filetypes)

    showinfo(
        title='Selected File',
        message=filename
    )


# open button
open_button = ttk.Button(
    root,
    text='Open a File',
    command=select_file
)

open_button.pack(expand=True)


# run the application
root.mainloop()


import unittest
import datetime
from geopack import geopack,t89,t96,t01,t04
import collections


def approx_eq(x,y, tolerance=1e-5):
    if isinstance(x, collections.Iterable):
        err = [abs(x0-y0) < tolerance for x0,y0 in zip(x,y)]
        return not (False in err)
    else:
        return abs(x-y) < tolerance


class KnownValues(unittest.TestCase):

    test = {
        'ut': (datetime.datetime(2001,1,1,2,3,4)-datetime.datetime(1970,1,1)).total_seconds(),
        'r_theta_phi': (1.1,0.1,0.2),
        'r_gsm': [-5.1,0.3,2.8],
        'v_gse': [-400,0,10],
        'r_gsw': [-5.1650466,0.31912656,2.6759018],
        'r_gse': (-5.1, 0.9056171572514691, 2.6664316163164146),
        'r_geo': (2.8011944117533565,-2.4048913761357267,4.5066403602406275),
        'r_sm': (-2.9670092644479498,0.3,5.004683409035982),
        'r_mag': (2.298686529948157,1.8997853069109853,5.004683409035982),
        'r_gei': (-0.05919908606124502,3.691434427381819,4.5066403602406275),
        'geod': (110,2),
        'geo': (6470.4815,0.431707561),
        'geod2': (109.546387,1.14329302),
        'geo2': (6470,0.43),
        'recalc': -0.53356312486214,
        'sun': (2.296253490174557,4.899558241818756,4.915948604659666,-0.40152585209539443,0.4090854263337441),
        'igrf': (262.8292494578462,-19.305779063359893,-50.34573331501855),
        'igrf_gsw': (263.86956,-20.599686,-43.992093),
        'dip': (266.04028284777775,-20.204186166677108,-57.492114467356956),
        'dip_gsw': (267.24898,-21.492340,-51.061152),
        't89': (20.77213175686351,-0.6465547428023687,-15.071641970338984),
        't96': (61.178346985193215,-1.4611972499959456,-40.44976904223083),
        't01': (46.35361850144623,1.4399149705997756,-31.998997670712665),
        't04': (10.251787900560483,-2.894661363001944,-9.615717866328614),
        'trace_setting': [-1,10,1.1],   # dir, rlim,r0 (Re).
        'trace_t89_igrf': (-0.7218581171377333,0.03278201781940832 ,0.8293646495541395),
        'trace_t96_igrf': (-0.7289210907095749,0.035354863396548725,0.8230575125223648),
        'trace_t01_igrf': (-0.726408232521053,0.03894789557265691  ,0.8251143626438783),
        'trace_t04_igrf': (-0.7153766644654653,0.024213930362513274,0.8352541020689167),
        'mgnp_shu': (-1.0470115138905702,1.4898498476086839,13.905265244347715),
        'mgnp_t96': (-1.140826561384304 ,1.4730846293697084,13.748789874117278),
        'bzimf': -5,
        'mgnp_t96_par': [10,500],   # density (cm^-3) ,velocity (km/s).
        'ps': -0.533585131
    }
    test['par0'] = [test['ps']]+test['r_gsm']
    test['par1'] = [2] + test['par0']
    test['par2'] = [[2,-87,2,-5, 0,0]+test['par0']] + test['par0']
    test['mgnp_shu_par'] = test['mgnp_t96_par']+ [test['bzimf']]

    def test_to_known_values(self):
        """geopack should give known result with known input"""

        # new recalc, which returns dipole tilt angle in rad.
        self.assertEqual(self.test['recalc'], geopack.recalc(self.test['ut'], *self.test['v_gse']))

        # new sun, which returns 5 angles in rad.
        self.assertEqual(self.test['sun'], geopack.sun(self.test['ut']))

        # new internal models.
        self.assertTrue(approx_eq(self.test['igrf'], geopack.igrf_gsm(*self.test['r_gsm'])))
        self.assertTrue(approx_eq(self.test['dip'], geopack.dip(*self.test['r_gsm'])))
        self.assertTrue(approx_eq(self.test['igrf_gsw'], geopack.igrf_gsw(*self.test['r_gsw']), 1e-3))
        self.assertTrue(approx_eq(self.test['dip_gsw'], geopack.dip_gsw(*self.test['r_gsw']), 1e-3))

        # new external models.
        self.assertTrue(approx_eq(self.test['t89'], t89.t89(*self.test['par1'])))
        self.assertTrue(approx_eq(self.test['t96'], t96.t96(*self.test['par2'])))
        self.assertTrue(approx_eq(self.test['t01'], t01.t01(*self.test['par2'])))
        self.assertTrue(approx_eq(self.test['t04'], t04.t04(*self.test['par2'])))

        # new coord transform, which returns B in nT.
        self.assertTrue(approx_eq(self.test['r_mag'], geopack.geomag(*self.test['r_geo'], 1)))
        self.assertTrue(approx_eq(self.test['r_geo'], geopack.geomag(*self.test['r_mag'], -1)))

        self.assertTrue(approx_eq(self.test['r_geo'], geopack.geigeo(*self.test['r_gei'], 1)))
        self.assertTrue(approx_eq(self.test['r_gei'], geopack.geigeo(*self.test['r_geo'], -1)))

        self.assertTrue(approx_eq(self.test['r_sm'],  geopack.magsm (*self.test['r_mag'], 1)))
        self.assertTrue(approx_eq(self.test['r_mag'], geopack.magsm (*self.test['r_sm'], -1)))

        self.assertTrue(approx_eq(self.test['r_gse'], geopack.gsmgse(*self.test['r_gsm'], 1)))
        self.assertTrue(approx_eq(self.test['r_gsm'], geopack.gsmgse(*self.test['r_gse'], -1)))

        self.assertTrue(approx_eq(self.test['r_gsm'], geopack.smgsm (*self.test['r_sm'], 1)))
        self.assertTrue(approx_eq(self.test['r_sm'],  geopack.smgsm (*self.test['r_gsm'], -1)))

        self.assertTrue(approx_eq(self.test['r_gsm'], geopack.geogsm(*self.test['r_geo'], 1)))
        self.assertTrue(approx_eq(self.test['r_geo'], geopack.geogsm(*self.test['r_gsm'], -1)))

        self.assertTrue(approx_eq(self.test['r_gsw'], geopack.gswgsm(*self.test['r_gsm'], -1)))
        self.assertTrue(approx_eq(self.test['r_gsm'], geopack.gswgsm(*self.test['r_gsw'], 1)))

        self.assertTrue(approx_eq(self.test['geo'], geopack.geodgeo(*self.test['geod'],  1), 1e-3))
        self.assertTrue(approx_eq(self.test['geod2'], geopack.geodgeo(*self.test['geo2'], -1), 1e-3))

        # new trace.
        self.assertTrue(approx_eq(self.test['trace_t89_igrf'],
                                  geopack.trace(*self.test['r_gsm'], *self.test['trace_setting'], self.test['par1'][0], 't89', 'igrf')))
        self.assertTrue(approx_eq(self.test['trace_t96_igrf'],
                                  geopack.trace(*self.test['r_gsm'], *self.test['trace_setting'], self.test['par2'][0], 't96', 'igrf')))
        self.assertTrue(approx_eq(self.test['trace_t01_igrf'],
                                  geopack.trace(*self.test['r_gsm'], *self.test['trace_setting'], self.test['par2'][0], 't01', 'igrf')))
        self.assertTrue(approx_eq(self.test['trace_t04_igrf'],
                                  geopack.trace(*self.test['r_gsm'], *self.test['trace_setting'], self.test['par2'][0], 't04', 'igrf')))

        # new magnetopaus model.
        self.assertTrue(approx_eq(self.test['mgnp_t96'], geopack.t96_mgnp(*self.test['mgnp_t96_par'], *self.test['r_gsm'])))
        self.assertTrue(approx_eq(self.test['mgnp_shu'], geopack.shuetal_mgnp(*self.test['mgnp_shu_par'], *self.test['r_gsm'])))


if __name__ == '__main__':
    unittest.main()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

from geospacelab.datahub import VariableModel as Var
import geospacelab.visualization.mpl.colormaps as cm

database = 'jhuapl'

timestamps = {
    'DATETIME': 'CENTERED_TIME',
}


default_colormap = cm.cmap_aurora()

default_plot_config = {
    'line':         {
        'linestyle':        '-',
        'linewidth':        1.5,
        'marker':           '',
        'markersize':       3,
    },
    'pcolormesh':   {
        'cmap':            default_colormap,
        'c_scale':         'log',
        'c_lim':           [100, 6000],
        'alpha':            0.9,
    }
}

configured_variables = {}
visual = 'on'

depend_0 = {'GEO_LAT': 'DISK_GEO_LAT', 'GEO_LON': 'DISK_GEO_LON'}
depend_1 = {'GEO_LAT': 'DISK_GEO_LAT'}
depend_2 = {'GEO_LAT': 'DISK_GEO_LAT', 'GEO_LON': 'DISK_GEO_LON'}
depend_c = {'SPECTRA': 'EMISSION_SPECTRA'}

####################################################################################################################
var_name = 'DISK_R_LBHS'
var = Var(name=var_name, ndim=3, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Auroral emission intensity at LBHS'
var.label = r'LBHS'
var.group = 'Emission intensity'
var.unit = 'R'
var.depends = {0: depend_0, 1: {'AACGM_LAT': 'GRID_MLAT'}, 2: {'AACGM_LON': 'GRID_MLON'}}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis


configured_variables[var_name] = var

configured_variables['DISK_R_RECT_LBHS'] = var.clone()

####################################################################################################################
var_name = 'DISK_R_LBHL'
var = Var(name=var_name, ndim=3, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Auroral emission intensity at LBHL'
var.label = r'LBHL'
var.group = 'Emission intensity'
var.unit = 'R'
var.depends = {0: depend_0, 1: {'AACGM_LAT': 'GRID_MLAT'}, 2: {'AACGM_LON': 'GRID_MLON'}}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis


configured_variables[var_name] = var

configured_variables['DISK_R_RECT_LBHL'] = var.clone()

####################################################################################################################
var_name = 'DISK_R_1304'
var = Var(name=var_name, ndim=3, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Auroral emission intensity at 130.4 nm'
var.label = r'1304'
var.group = 'Emission intensity'
var.unit = 'R'
var.depends = {0: depend_0, 1: {'AACGM_LAT': 'GRID_MLAT'}, 2: {'AACGM_LON': 'GRID_MLON'}}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis


configured_variables[var_name] = var

configured_variables['DISK_R_RECT_1304'] = var.clone()

####################################################################################################################
var_name = 'DISK_R_1356'
var = Var(name=var_name, ndim=3, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Auroral emission intensity at 135.6 nm'
var.label = r'1356'
var.group = 'Emission intensity'
var.unit = 'R'
var.depends = {0: depend_0, 1: {'AACGM_LAT': 'GRID_MLAT'}, 2: {'AACGM_LON': 'GRID_MLON'}}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis


configured_variables[var_name] = var

configured_variables['DISK_R_RECT_1356'] = var.clone()

####################################################################################################################
var_name = 'DISK_R_1216'
var = Var(name=var_name, ndim=3, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Auroral emission intensity at 121.6 nm'
var.label = r'1216'
var.group = 'Emission intensity'
var.unit = 'R'
var.depends = {0: depend_0, 1: {'AACGM_LAT': 'GRID_MLAT'}, 2: {'AACGM_LON': 'GRID_MLON'}}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis


configured_variables[var_name] = var

configured_variables['DISK_R_RECT_1216'] = var.clone()